// "El Dumbo" and El Dumbo 2 scalable cipher modes
// Basic function is 64 bit block/128 bit key (IDEA, XTEA..)
// Size N is integer number of 64 bit blocks ( >= 1)
// Key size is 2 x N
//

#include "types.h"
#include <stdio.h>
#include "el_dumbo.h"
#include "string.h"


void El_Dumbo_Encode(void (*encode)(u32 *x, u32 const* key),  u32 *x, u32 *const key, const u32 size_x64)
{

if(size_x64 == 1)
 {
 encode(x, key);
 return;
 }

u32 k[4];
u32 tmp[2];
u32 cnt = 0;
u32 rounds = 2*size_x64;

while(rounds--)
 {
 u32 j = 0;

 for(u32 i = 0; i < (2*size_x64 - 1); i++)
  {
  k[0] = key[j + 0];
  k[1] = key[j + 1];
  k[2] = key[j + 2];
  k[3] = key[j + 3];

  if(j == (4*size_x64 - 4)) j = 0;
  else j += 4;

  k[0] ^= cnt;
  cnt++;
  encode(x + i, k);
  }

 k[0] = key[j + 0];
 k[1] = key[j + 1];
 k[2] = key[j + 2];
 k[3] = key[j + 3];
 
 k[0] ^= cnt;
 cnt++;

 tmp[0] = x[2*size_x64 - 1];
 tmp[1] = x[0];
 encode(tmp, k);
 x[2*size_x64 - 1] = tmp[0];
 x[0] = tmp[1];
 }
}
//-------------------------------------------------------------

void El_Dumbo_Decode(void (*decode)(u32 *x, u32 const* key),  u32 *x, u32 *const key, const u32 size_x64)
{

if(size_x64 == 1)
 {
 decode(x, key);
 return;
 }

u32 k[4];
u32 tmp[2];
u32 rounds = 2*size_x64;
u32 cnt = rounds*2*size_x64;

while(rounds--)
 {
 u32 j = 4*size_x64 - 4;

 k[0] = key[j + 0];
 k[1] = key[j + 1];
 k[2] = key[j + 2];
 k[3] = key[j + 3];

 cnt--;
 k[0] ^= cnt;
 
 tmp[0] = x[2*size_x64 - 1];
 tmp[1] = x[0];
 decode(tmp, k);
 x[2*size_x64 - 1] = tmp[0];
 x[0] = tmp[1];

 for(s32 i = (2*size_x64 - 2); i >= 0; i--)
  {
  if(j) j -= 4;
  else  j  = 4*size_x64 - 4;

  k[0] = key[j + 0];
  k[1] = key[j + 1];
  k[2] = key[j + 2];
  k[3] = key[j + 3];
  
  cnt--;
  k[0] ^= cnt;
  decode(x + i, k);
  }
 }
}
//-------------------------------------------------------------------
// 2 x 2n transposition
//
static void dumbo_transpose(u32 *x, u32 *tmp, u32 len)
{
for(u32 i = 0; i < len/2; i++)
 {
 tmp[2*i + 0] = x[i + 0];
 tmp[2*i + 1] = x[i + len/2];
 }

for(u32 i = 0; i < len; i++) x[i] = tmp[i];
}

static void dumbo_inv_transpose(u32 *x, u32 *tmp, u32 len)
{
for(u32 i = 0; i < len/2; i++)
 {
 tmp[i] = x[2*i + 0];
 tmp[i + len/2] = x[2*i + 1];
 }
for(u32 i = 0; i < len; i++) x[i] = tmp[i];
}

static u32 dumbo_rounds(const u32 size_x64)
{
// Find minimum rounds to achieve dependency of all blocks and all keys

/*
u32 diffusion_rounds = 1;
u32 mask  = 1;
while(size_x64 > mask) // log2(N)
 {
 mask <<= 1;
 diffusion_rounds++;
 }
u32 rounds = diffusion_rounds + size_x64;
return rounds;
*/
// diffusion rounds are always less or equal to the number of keys
// Therefore it is suffucient to run through N rounds to achieve full dependency of blocks and keys
// Note it does not produce mixing adequate to increased block size


return size_x64;
}

//----------- Using transpose --------------------
// tmp array of the size of same as data
//
void El_Dumbo_2_Encode(void (*encode)(u32 *x, u32 const* key),  u32 *x, u32 *tmp, u32 *const key, const u32 size_x64)
{

if(size_x64 == 1)
 {
 encode(x, key);
 return;
 }


u32 rounds = dumbo_rounds(size_x64);

u32 k[4];
u32 k_idx = 0;
u32 cnt = 0;

for(u32 j = 0; j < rounds; j++)
 {
 u32 const* kptr = key + 4*k_idx;
 k_idx++;
 if(k_idx >= size_x64) k_idx = 0;

 k[1] = kptr[1];
 k[2] = kptr[2];
 k[3] = kptr[3];

 for(u32 i = 0; i < size_x64; i++)
  {
  k[0] = kptr[0];
  k[0] ^= cnt;
  cnt++;
  encode(x + 2*i, k);
  } //i
 if(j < (rounds - 1)) dumbo_transpose(x, tmp, 2*size_x64); // Do not transpose last round
 } // j

// Cleanup
for(u32 i = 0; i < 2*size_x64; i++) tmp[i] = 0;
}

// tmp array of the size of same as data

void El_Dumbo_2_Decode(void (*decode)(u32 *x, u32 const* key),  u32 *x, u32 *tmp, u32 *const key, const u32 size_x64)
{

if(size_x64 == 1)
 {
 decode(x, key);
 return;
 }


s32 rounds = dumbo_rounds(size_x64);

u32 k[4];
u32 k_idx = 4*((rounds - 1) % size_x64);
u32 cnt = rounds * size_x64;

for(s32 j = rounds - 1; j >= 0; j--)
 {

 if(j < (rounds - 1)) dumbo_inv_transpose(x, tmp, 2*size_x64); // Do not transpose last round

 u32 const* kptr = key + k_idx;
 if(k_idx) k_idx -= 4;
 else k_idx = 4*(size_x64 - 1);

 k[1] = kptr[1];
 k[2] = kptr[2];
 k[3] = kptr[3];


 for(s32 i = size_x64 - 1; i >= 0; i--)
  {
  k[0] = kptr[0];
  cnt--;
  k[0] ^= cnt;
  decode(x + 2*i, k);
  } // i
 } // j

// Cleanup
for(u32 i = 0; i < 2*size_x64; i++) tmp[i] = 0;
}
//------------------
// Keyless permutation
//
void El_Dumbo_3_Encode(void (*encode)(u32 *x, u32 const* key),  u32 *x,  const u32 size_x64, const u32 r)
{
u32 key[4];



u32 rounds = r;
if(!rounds) rounds = size_x64;

u32 N = 0;

while(rounds--)
 {
 u32 j = 1;
 u32 k = size_x64 - 1;

 for(u32 i = 0; i < size_x64; i++)
  {
  key[0] = x[2*j + 0];
  key[1] = x[2*j + 1];
  j++;
  if(j == size_x64) j = 0;

  key[2] = x[2*k + 0];
  key[3] = x[2*k + 1];
  k++;
  if(k == size_x64) k = 0;

  x[2*i + 0] ^= N;
  N++;
  encode(x + 2*i, key);
  }
 }
}

void El_Dumbo_3_Decode(void (*decode)(u32 *x, u32 const* key),  u32 *x,  const u32 size_x64, const u32 r)
{
u32 key[4];

u32 rounds = r;
if(!rounds) rounds = size_x64;

u32 N = rounds*size_x64  - 1;

while(rounds--)
 {
 u32 j = 0;
 u32 k = size_x64 - 2;

 for(s32 i = (size_x64 - 1); i >= 0; i--)
  {
  key[0] = x[2*j + 0];
  key[1] = x[2*j + 1];
  if(j) j--;
  else j = size_x64 - 1;
  
  key[2] = x[2*k + 0];
  key[3] = x[2*k + 1];
  if(k) k--;
  else k = size_x64 - 1;

  decode(x + 2*i, key);

  x[2*i + 0] ^= N;
  N--;
  }
 }
}


//=================================== <END> =========================

