#ifndef ENCRYPT_DECRYPT_PACKET_H_INCLUDED
#define ENCRYPT_DECRYPT_PACKET_H_INCLUDED
#include "random.h"
#include "keystream.h"

static const u32 max_data_length = 32768; // Data field can't be longer then this
//---------------------------------------
// Small encrypt (minimize overhead)
// 
static const u16 small_encrypt_packet_overhead_bytes = 6;  // 2 byte cntr + 4 bytes MIC
static const u16 small_encrypt_max_data_size = 65535 - small_encrypt_packet_overhead_bytes; 
static const u16 small_encrypt_max_skip_forward_bytes = 49152; // If the counter jumped forward more then this, then packet is invalid (or from the past)


class ENCRYPT_DECRYPT_PACKET
{
KEYSTREAM_GENERATOR keystream; // keystream generator 
KEYSTREAM_GENERATOR tmp_keystream_state;

u32 mic_iv[2];  // 64 bit IV for mic
u32 mic_key[4]; // 128-bit key for mic

u32 max_data_len;

u64 invalid_packets_cntr;
u64 late_packets_cntr;

u32 small_compute_mic(u8 const *data, const u16 data_len_bytes); // Length excluding mic


  
public:
     ENCRYPT_DECRYPT_PACKET(u32 packet_max_data_length_bytes = max_data_length);
     ~ENCRYPT_DECRYPT_PACKET();
     void reset();
     void init(u32 const *key_data); // Initialize encryption/decryption with u32[18] key
     u64 total_rx_bytes();
     u64 dropped_rx_bytes();
     u64 invalid_packets(); // Count of invalid RX packets (wrong format or wrong MIC)
     u64 late_packets(); // Count of late RX packets from the past
     u16 small_encrypt(u8 *dest, u8 const *src, const u16 src_len_bytes);  // Packet encryption, returns length of encrypted block
     u16 small_decrypt(u8 *dest, u8 const *src, const u16 src_len_bytes);  // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)
     
                
};

#endif

