#ifndef FAST_MODULO_H_INCLUDED
#define FAST_MODULO_H_INCLUDED

#include "types.h"
#include "fast_mul.h"

class FAST_MODULO
{
u32 len;
u32 e;
u32 *p;
u32 *inverse_p;
u32 *tmp0;
u32 *tmp1;
u32 *tmp2;
u32 *tmp_prod;
FAST_MUL *fast_mul; 

public:
    FAST_MODULO(u32 const length_x32); // length of P in 32-bit words
    ~FAST_MODULO();
    void reset();
    void fast_modulo_init(u32 const* p);  
    void fast_modulo(u32 *result, u32 const * x); // x modulo p; x < p^2, the size of result is length_x32  
                                                  // size of x = 2 x length_x32
};


class FAST_MULTI_POW_MOD 
{
u32 length;
u32 *pwr_x;
u32 *tmp_prod;

public:
    FAST_MODULO *fm;      // made it public; could be used further in MRT
    FAST_MUL *fast_mul;   // made it public; could be used further in MRT
 
    FAST_MULTI_POW_MOD(u32 length_x32);
    ~FAST_MULTI_POW_MOD();
    void multi_pow_mod(u32 *result, const u32 *x, const u32 *y, const u32 *p);   // (x ^ y) modulo p
};


#endif

