#ifndef HASH_FUNC_H_INCLUDED
#define HASH_FUNC_H_INCLUDED


#include "types.h"
#include "sc128.h"
#include "el_dumbo.h"


void Hash_Block_X_64(u32 *hash, u8 const* data, u32 data_length_bytes); // 64 bit hash, any data length

//-------------------------
// Universal hash function
// Concept similar to SKEIN, however internals are different
// Speed doesn't matter as DH computation takes ~1000 times more cpu.
// Also, it would be difficult to implement this function in hardware

class COLLAPSE
 {
 u32 size_x32;
 u32 size_x128;
 u32 data_byte_cnt;
 u64 tweak;
 
 SC128_CIPHER *sc128;


 u64 *data_input;
 u64 *lock_bar;
 u64 *hash_input;

 u32 *el_dumbo_tmp;

 u64 prng();
 void hash_primitive();

    
public:
    COLLAPSE(u32 hash_size_x32);
    ~COLLAPSE();

    void init_hash(const u32 data_length_bytes, const u32 iv = 0);
    void process_hash(const u8 data);
    void process_hash(void const* data, u32 data_length_bytes);
    void complete_hash(u32 *output);

    // Hash a block of data
    void hash(u32 *hash, void const *data, u32 data_block_size_bytes, const u32 iv = 0);

    // Can be used as block cipher for testing

    void encrypt(u64 *data, u64 const* key);
    void decrypt(u64 *data, u64 const *key);

    void reset();
 };

//----------------------------------------
//
class DH_HASH
 {
 COLLAPSE *collapse;
       
public:
    DH_HASH(u32 size_x32);
    ~DH_HASH();
    void hash(u32 *hash, u32 const* g, u32 const* y, u32 const* p, const u32 dh_length_x32);
    void hash(u32 *hash, u32 const* g, u32 const *p, u32 const* gxp, u32 const* gyp, const u32 dh_length_x32);
    void reset();
        
 };


int collapse_test();


#endif

