#ifndef KEYSTREAM_H_INCLUDED
#define KEYSTREAM_H_INCLUDED

#include "types.h"

void keystream_generator(u8 *output, u32 *state);
void keystream_generator_init(u32 *state, u32 const *key);

const u32 keystream_state_size_x32 = 12;
const u32 keystream_state_size_x64 = 6;

// No dynamic allocations in this class
class KEYSTREAM_GENERATOR
{
u8 byte_buffer[16];
u32 byte_buffer_idx;
u32 state[keystream_state_size_x32];
u64 cntr;
u64 skip_cntr;



public:
    KEYSTREAM_GENERATOR();
    ~KEYSTREAM_GENERATOR();
    void init_key(u32 const* key); // Initalize with 384-bit key u32[12]    
    u8 get_keystream_byte(); // Generate pseudorandom byte
    int counter(u64 cnt); // returns 0 if counter less then current (can't go back in time)
    u64 counter(); // returns counter
    u64 skip_counter(); // returns skipped counter
    void fast_forward(u64 fwd_bytes); // fast forward by X bytes
    void reset();  

};

int test_keystream();


#endif

