#ifdef _LINUX
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <fnmatch.h>
#include <unistd.h>
#elif defined(_WIN32)
#include <windows.h>
#endif
//---------------------------------------------------------------------------
#include "cp_fileop.h"
#include "cp_strop.h"
//---------------------------------------------------------------------------
fd_t cp_open(pctstr_t fname,int mode)
{
#ifdef _LINUX
	return open(fname,mode,0644);
#elif defined(_WIN32)
	DWORD access = 0,create = OPEN_EXISTING,attr = 0;
	if (mode & FO_RDONLY) access |= GENERIC_READ;
	if (mode & FO_WRONLY) {
		access |= GENERIC_WRITE;
		create = OPEN_ALWAYS;
	}
	if (mode & FO_CREAT) {
		create = CREATE_ALWAYS;
		attr = FILE_ATTRIBUTE_NORMAL;
	}
	if (mode & FO_TRUNC)
		create = TRUNCATE_EXISTING;

	return CreateFile(fname,access,0,NULL,create,attr,NULL);
#endif
}
//---------------------------------------------------------------------------
int cp_read(fd_t file,void *data,u32 szdata)
{
#ifdef _LINUX
	return read(file,data,szdata);
#elif defined(_WIN32)
	DWORD dwRead;

	if (!ReadFile(file,data,szdata,&dwRead,0))
		return -1;

	return (int)dwRead;
#endif
}
//---------------------------------------------------------------------------
int cp_write(fd_t file,const void *data,u32 szdata)
{
#ifdef _LINUX
	return write(file,data,szdata);
#elif defined(_WIN32)
	DWORD dwWrite;

	if (!WriteFile(file,data,szdata,&dwWrite,0))
		return -1;
	return (int)dwWrite;
#endif
}
//---------------------------------------------------------------------------
int cp_close(fd_t file)
{
#ifdef _LINUX
	return close(file);
#elif defined(_WIN32)
	return CloseHandle(file);
#endif
}
//---------------------------------------------------------------------------
#ifdef _WIN32
static void FT2TT(FILETIME *ft,time_t *tt)
{ *tt = (((LONGLONG)ft->dwHighDateTime << 32) + ft->dwLowDateTime - 116444736000000000LL)/10000000; }
#endif
//---------------------------------------------------------------------------
int cp_scandir(pctstr_t dir,pctstr_t pattern,fileinfo_t **result)
{
	int n = 0;
#ifdef _LINUX
	struct dirent **nl = 0;
	struct stat st;
	int sr,i;
	*result = 0;

	if ((sr = scandir(dir,&nl,0,alphasort))) {
		for (i = 0; i < sr; i++) {
			if (!fnmatch(pattern,nl[i]->d_name,FNM_NOESCAPE)) {
				*result = (fileinfo_t*)realloc(*result,++n*sizeof(fileinfo_t));
				(*result)[n-1].flags = 0;

				if (!stat(nl[i]->d_name,&st)) {
					if (S_ISDIR(st.st_mode)) (*result)[n-1].flags |= FF_DIR;
					if (S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode)) (*result)[n-1].flags |= FF_DEVICE;
					if (S_ISLNK(st.st_mode)) (*result)[n-1].flags |= FF_LINK;
					(*result)[n-1].size = st.st_size;
					(*result)[n-1].atime = st.st_atime;
					(*result)[n-1].ctime = st.st_ctime;
					(*result)[n-1].mtime = st.st_mtime;
				}
				cp_snprintf((*result)[n-1].name,MAXFILENAME,"%s/%s",dir,nl[i]->d_name);
			}
			free(nl[i]);
		}
		free(nl);
	}
#elif defined(_WIN32)
	WIN32_FIND_DATA fd;
	HANDLE hFind;
	tchar_t path[MAXFILENAME];

	if (!result) return -1;

	cp_strcbcpy(path,sizeof(path),dir);
	if (path[cp_strlen(path)-1] != '\\')
		cp_strcbcat(path,sizeof(path),_T("\\"));
	cp_strcbcat(path,sizeof(path),pattern);

	*result = 0;
	if ((hFind = FindFirstFile(path,&fd)) != INVALID_HANDLE_VALUE) {
		do {
			fileinfo_t *fi = (fileinfo_t*)realloc(*result,++n*sizeof(fileinfo_t));
			if (!fi) break;

			*result = fi;
			(*result)[n-1].flags = 0;
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) (*result)[n-1].flags |= FF_DIR;
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DEVICE) (*result)[n-1].flags |= FF_DEVICE;
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) (*result)[n-1].flags |= FF_LINK;
			cp_snprintf((*result)[n-1].name,_CCH((*result)[n-1].name),_T("%s\\%s"),dir,fd.cFileName);
			(*result)[n-1].size = ((u64)fd.nFileSizeHigh << 32) | fd.nFileSizeLow;
			FT2TT(&fd.ftCreationTime,&(*result)[n-1].ctime);
			FT2TT(&fd.ftLastWriteTime,&(*result)[n-1].mtime);
			FT2TT(&fd.ftLastAccessTime,&(*result)[n-1].atime);
		} while (FindNextFile(hFind,&fd));
		FindClose(hFind);
	}
#endif

	return n;
}
//---------------------------------------------------------------------------
int cp_subdirs(pctstr_t dir,fileinfo_t **result)
{
	int n = 0;
#ifdef _LINUX
	struct dirent **nl = 0;
	struct stat st;
	int sr,i;
	*result = 0;

	if ((sr = scandir(dir,&nl,0,alphasort))) {
		for (i = 0; i < sr; i++) {
			if (!stat(nl[i]->d_name,&st) && S_ISDIR(st.st_mode) &&
				cp_strcmp(nl[i]->d_name,_T(".")) && cp_strcmp(nl[i]->d_name,_T("..")))
			{
				fileinfo_t *fi = (fileinfo_t*)realloc(*result,++n*sizeof(fileinfo_t));
				if (!fi) break;

				*result = fi;

				if (S_ISLNK(st.st_mode)) (*result)[n-1].flags |= FF_LINK;
				(*result)[n-1].size = st.st_size;
				(*result)[n-1].atime = st.st_atime;
				(*result)[n-1].ctime = st.st_ctime;
				(*result)[n-1].mtime = st.st_mtime;
			}
			cp_snprintf((*result)[n-1].name,MAXFILENAME,"%s/%s",dir,nl[i]->d_name);

			free(nl[i]);
		}
		free(nl);
	}
#elif defined(_WIN32)
	WIN32_FIND_DATA fd;
	HANDLE hFind;
	tchar_t path[MAXFILENAME];

	if (!result) return -1;

	cp_strcbcpy(path,sizeof(path),dir);
	if (path[cp_strlen(path)-1] == '\\')
		cp_strcbcat(path,sizeof(path),_T("*"));
	else
		cp_strcbcat(path,sizeof(path),_T("\\*"));

	*result = 0;
	if ((hFind = FindFirstFileEx(path,FindExInfoBasic,&fd,FindExSearchLimitToDirectories,0,FIND_FIRST_EX_LARGE_FETCH)) != INVALID_HANDLE_VALUE) {
		do
		if ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
			cp_strcmp(fd.cFileName,_T(".")) && cp_strcmp(fd.cFileName,_T("..")))
		{
			fileinfo_t *fi = (fileinfo_t*)realloc(*result,++n*sizeof(fileinfo_t));
			if (!fi) break;

			*result = fi;
			(*result)[n-1].flags = 0;
			cp_snprintf((*result)[n-1].name,_CCH((*result)[n-1].name),_T("%s\\%s"),dir,fd.cFileName);
			(*result)[n-1].size = ((u64)fd.nFileSizeHigh << 32) | fd.nFileSizeLow;
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) (*result)[n-1].flags |= FF_LINK;
			FT2TT(&fd.ftCreationTime,&(*result)[n-1].ctime);
			FT2TT(&fd.ftLastWriteTime,&(*result)[n-1].mtime);
			FT2TT(&fd.ftLastAccessTime,&(*result)[n-1].atime);
		} while (FindNextFile(hFind,&fd));
		FindClose(hFind);
	}
#endif

	return n;
}
//---------------------------------------------------------------------------
s64 cp_seek(fd_t file,s64 offs,int wh)
{
#ifdef _LINUX
	return lseek64(file,offs,wh);
#elif defined(_WIN32)
	LARGE_INTEGER md,no;
	md.QuadPart = offs;
	if (!SetFilePointerEx((HANDLE)file,md,&no,wh))
		return (s64)-1;
	return no.QuadPart;
#endif
}
//---------------------------------------------------------------------------
