#ifndef CP_THREADS_H_INCLUDED
#define CP_THREADS_H_INCLUDED
//---------------------------------------------------------------------------
#if !defined(_WIN32) && !defined(_LINUX)
#error "OS not specified!"
#endif
//---------------------------------------------------------------------------
#ifdef _LINUX
#include <pthread.h>
#include <unistd.h>

typedef pthread_t thread_t;
typedef pthread_mutex_t lock_t;
typedef pthread_cond_t cond_t;

#define TP_IDLE					 15
#define TP_LOWEST				 10
#define TP_LOW					 5
#define TP_NORMAL				 0
#define TP_HIGH					-5
#define TP_HIGHEST				-10
#define TP_CRITICAL				-15

#define cp_thread_set_priority	pthread_setschedprio

#define msleep(v)				usleep(v*1000)

#define cp_initlock(lock)		pthread_mutex_init(lock,0)
#define cp_deletelock			pthread_mutex_destroy
#define cp_lock					pthread_mutex_lock
#define cp_trylock				pthread_mutex_trylock
#define cp_unlock				pthread_mutex_unlock

#define cp_initcond(cond)		pthread_cond_init(cond,0)
#define cp_deletecond			pthread_cond_destroy
#define cp_waitcond				pthread_cond_wait
#define cp_wakecond				pthread_cond_signal
#define cp_wakeallcond			pthread_cond_broadcast

#endif
//---------------------------------------------------------------------------
#ifdef _WIN32
#include <windows.h>

typedef HANDLE thread_t;
typedef PCONDITION_VARIABLE cond_t;

#define TP_IDLE					THREAD_PRIORITY_IDLE
#define TP_LOWEST				THREAD_PRIORITY_LOWEST
#define TP_LOW					THREAD_PRIORITY_BELOW_NORMAL
#define TP_NORMAL				THREAD_PRIORITY_NORMAL
#define TP_HIGH					THREAD_PRIORITY_ABOVE_NORMAL
#define TP_HIGHEST				THREAD_PRIORITY_HIGHEST
#define TP_CRITICAL				THREAD_PRIORITY_TIME_CRITICAL

#define cp_thread_set_priority	SetThreadPriority

#define sleep(v)				Sleep(v*1000)
#define msleep(v)				Sleep(v)
#define usleep(v)				Sleep(v/1000)

#define cp_initcond				InitializeConditionVariable
#define cp_deletecond(cond)
#define cp_wakecond				WakeConditionVariable
#define cp_wakeallcond			WakeAllConditionVariable
#define cp_waitcond(cond,lock)	SleepConditionVariableCS(cond,lock,INFINITE)

typedef CRITICAL_SECTION lock_t;

#define cp_initlock				InitializeCriticalSection
#define cp_deletelock			DeleteCriticalSection
#define cp_lock					EnterCriticalSection
#define cp_trylock				TryEnterCriticalSection
#define cp_unlock				LeaveCriticalSection

#endif
//---------------------------------------------------------------------------
#define THREAD_ERROR	(thread_t)0
//---------------------------------------------------------------------------
typedef int (*threadproc_t)(void*);
//---------------------------------------------------------------------------
thread_t cp_thread_create(threadproc_t proc,void *param);
int cp_thread_close(thread_t thread);
//---------------------------------------------------------------------------
#endif //CP_THREADS_H_INCLUDED
