#include <winsock2.h>
#include <Mswsock.h>
#include "network.h"
#include "password.h"
#include "resource.h"
#include "log.h"
#include "hash_func.h"
#include "random.h"
#include "diffie_helman.h"
#include "main.h"
#include "build.h"
#include "obfuscation.h"
#include "upnp.h"
#include <tchar.h>
#include <CommCtrl.h>
#include <Windows.h>
//-----------------------------------------
static LPTSTR module = TEXT("Network");
//-----------------------------------------
bool isIPLocal(u32 ip)
{
	if ((ip>>24) == 10 || // 10.x.x.x
		(ip>>16) == ((192<<8)|168) || // 192.168.x.x
		(ip>>20) == ((172<<4)|1)) // 172.16.x.x - 172.31.x.x
		return true;
	return false;
}
//-----------------------------------------
ESTABLISHMENT::ESTABLISHMENT()
{
	replied = false;
	state = NET_STATE_IDLE;
	timeout = 0;
	send_requests = 0;
	rec_flags = 0;
	rec_size = 0;
	last_size = 0;
	sent_size = 0;
	conf_size = 0;
	ip = 0;
	port = 0;
	local_ip = 0;
	local_port = 0;
	protocol = 0;
	version = 0;
	ZeroMemory(encapsulated_id,sizeof(encapsulated_id));
}
//-----------------------------------------
ESTABLISHMENT::~ESTABLISHMENT()
{
	tx_session.reset();
	rx_session.reset();
}
//-----------------------------------------

CONNECTION::CONNECTION(const u32* id, u32 ip, u16 port)
{
	CopyMemory(&nd.id,id,NETWORK_ID_LENGTH_BYTES);
	nd.ip = ip;
	nd.port = port;
	established = 0;
	next_connection = NULL;
	InitializeCriticalSection(&est_cs);
	est = NULL;
	ZeroMemory(encapsulated_id,sizeof(encapsulated_id));
	pack_num_est = 0;
	pack_num_sec = 0;
	pack_num_drop = 0;
	pack_num_sent = 0;
	send_cnt = 0;
	away = false;
	cp_strcchcpy((tchar_t*)userInfo,sizeof(userInfo)/sizeof(tchar_t),_T("<unknown>"));
	userInfoSize = (u32)cp_strlen((tchar_t*)userInfo)*sizeof(tchar_t);
}
//-----------------------------------------
CONNECTION::~CONNECTION()
{
	EnterCriticalSection(&est_cs);
	if (est) delete est;
	est = NULL;
	LeaveCriticalSection(&est_cs);
	DeleteCriticalSection(&est_cs);
	ses.reset();
}
//-----------------------------------------------------------------------
DWORD WINAPI receiverProc(NETWORK* n);
DWORD WINAPI timeoutProc(NETWORK* n);
//-----------------------------------------------------------------------
NETWORK::NETWORK(NODE* me, HWND mWnd)
{
	InitializeCriticalSection(&cs);
	con = NULL;
	away = false;
	session_is_initialized = false;

	CopyMemory(my_id,me->id,NETWORK_ID_LENGTH_BYTES);

    mainWnd = mWnd;
	recThread = NULL;

	// open new socket
	soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
	if (soc == INVALID_SOCKET)
	{
		MessageBox(mWnd,TEXT("Can't create socket"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
	ZeroMemory(&enc_host,sizeof(enc_host));
	ZeroMemory(&my_addr,sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = INADDR_ANY;
	my_addr.sin_port = htons(me->port);
	local_ip = getInternIp();
	local_port = me->port;
	if (SOCKET_ERROR == bind(soc,(sockaddr*)&my_addr,sizeof(my_addr)))
	{
		MessageBox(mWnd,TEXT("Can't bind socket to local port"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
    // set receiver timeout in ms
	DWORD timeout = 100;
	setsockopt(soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
	// disable  new behavior using
	// IOCTL: SIO_UDP_CONNRESET
	DWORD dwBytesReturned = 0;
	BOOL bNewBehavior = FALSE;
	WSAIoctl(soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
	// Initialize sending packet
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (pas_len)
	{
		// create receiver thread and stop event
		eStopThreads = CreateEvent(NULL,TRUE,FALSE,NULL);
		recThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)receiverProc,this,0,NULL);
		tmoThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)timeoutProc, this, 0, NULL);
	}
	else
	{
		eStopThreads = NULL;
		recThread = NULL;
		tmoThread = NULL;
	}
}
//-----------------------------------------------------------------------
NETWORK::~NETWORK()
{
	if (recThread)
	{
		SetEvent(eStopThreads);
		WaitForSingleObject(recThread,INFINITE);
		CloseHandle(recThread);
		WaitForSingleObject(tmoThread,INFINITE);
		CloseHandle(tmoThread);
	}
	if (con) // clear all connections
	{
		u32 term = SESSION_TERMINATE|(away?SESSION_USER_AWAY:0);
		while(con)
		{
			CONNECTION* cn = con->next_connection;
			if (con->established)
			{
				send_small_buffer(con,term,NULL,0);
			}
			delete con;
			con = cn;
		}
	}
	closesocket(soc);
    DeleteCriticalSection(&cs);
	cmn_tx_session.reset();
}
//-----------------------------------------------------------------------
CONNECTION* NETWORK::find_connection(const u32* id)
{
	for (CONNECTION *s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(s->nd.id,id))
		{
			return s;
		}
	}
	return NULL;
}
//-----------------------------------------------------------------------
CONNECTION* NETWORK::find_connection(u32 ip, u16 port)
{
	for (CONNECTION *s = con; s; s = s->next_connection)
	{
		if ((s->nd.ip == ip) && (s->nd.port == port))
		{
			return s;
		}
	}
	for (CONNECTION *s = con; s; s = s->next_connection)
	{
		if ((s->nd.local_ip == ip) && (s->nd.local_port == port))
		{
			return s;
		}
	}
	return NULL;
}
//-----------------------------------------------------------------------
CONNECTION* NETWORK::find_connection(const u32* id, u32 ip, u16 port, bool establish)
{
	u32 count = 0;
	CONNECTION *s, *prev_s;
	if (!con) // there is no any session
	{
		if (establish)
		{
			con = new CONNECTION(id,ip,port);
			return con;
		}
		else
		{
			return NULL;
		}
	}
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(s->nd.id,id))
		{
			return s;
		}
		prev_s = s;
		count++;
	}
	if (!establish) return NULL;
	if (count < NETWORK_MAX_CONNECTION_NUMBER)
	{
		prev_s->next_connection = new CONNECTION(id,ip,port);
		return prev_s->next_connection;
	}
	else
	{
		for (s = con; s; s = s->next_connection)
		{
			if (!s->established)
			{
				EnterCriticalSection(&s->est_cs);
				memcpy(s->nd.id,id,NETWORK_ID_LENGTH_BYTES);
				s->nd.ip = ip;
				s->nd.port = port;
				s->pack_num_drop = 0;
				s->pack_num_est = 0;
				s->pack_num_sent = 0;
				s->pack_num_sec = 0;
				if (s->est) delete s->est;
				s->est = NULL;
				LeaveCriticalSection(&s->est_cs);
				return s;
			}
		}
	}
	return NULL;
}
//-----------------------------------------------------------------------
#define OBFUSCATION
static void sendto_obfuscation(SOCKET soc,const char* data, int len, int flags,const sockaddr* addr,int tolen)
{
#ifdef OBFUSCATION
	char obf_data[NETWORK_SIZE_UDP];
	u32 size = Small_Obfuscate_Packet((u8*)obf_data, (u8*)data, len, IsObfuscationEnabled());
	sendto(soc,obf_data, size, flags,addr,tolen);
#else
	sendto(soc,data, len, flags,addr,tolen);
#endif
}
//-----------------------------------------------------------------------
static int recvfrom_obfuscation(SOCKET soc,char* data, int len, int flags,sockaddr* addr,int *fromlen)
{
#ifdef OBFUSCATION
	char obf_data[NETWORK_SIZE_UDP];
	len = recvfrom(soc,obf_data,len,flags,addr,fromlen);
	if (len > 0)
	{
		len = Small_Deobfuscate_Packet((u8*)data, (u8*)obf_data, len);
	}
#else
	len = recvfrom(soc,data,len,flags,addr,fromlen);
#endif
	return len;
}
//-----------------------------------------------------------------------
void NETWORK::send_to_host(CONNECTION* s, CONNECTION* host, const u8* buf, u32 size)
{
	/*
	u32 *data = new u32[size/sizeof(u32)+NETWORK_ID_LENGTH_WORDS+1+64];
	if (!data) return;
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(host->nd.ip);
	addr.sin_port = htons(host->nd.port);
	memcpy(data,my_id,NETWORK_ID_LENGTH_BYTES);
	data[NETWORK_ID_LENGTH_WORDS] = SESSION_ENC_FOR|(away?SESSION_USER_AWAY:0);
	size = host->ses.encrypt((u8*)&data[NETWORK_ID_LENGTH_WORDS+1],(u8*)buf,size);
	LogLowLevel(true,ntohl(addr.sin_addr.s_addr),ntohs(addr.sin_port),s->nd.id,buf[NETWORK_ID_LENGTH_WORDS]&0xFF,size);
//	sendto(soc,(char*)data,size+(NETWORK_ID_LENGTH_WORDS+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
	sendto_obfuscation(soc,(char*)data,size+(NETWORK_ID_LENGTH_WORDS+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
	delete[] data;
	*/
}
//-----------------------------------------------------------------------
void NETWORK::send_buf(CONNECTION* s, bool est, u8* buf, u32 size)
{
	CONNECTION* host;

	if (est && !s->est) return;

	if ((!est && isZeroID(s->encapsulated_id)) || (est && isZeroID(s->est->encapsulated_id))) // normal packet
	{
		if (encapsulation() && isNonZeroID(enc_host) && (host = find_connection(enc_host)) && host->established && !isIdIdentical(host->nd.id,s->nd.id))
		{
			//memcpy(buf,s->nd.id,NETWORK_ID_LENGTH_BYTES);
			//send_to_host(s,host,buf,size);
		}
		else // regular direct packet
		{
			sockaddr_in addr = {0};
			addr.sin_family = AF_INET;
			if (!est)
			{
				bool local = (getExternIp() == s->nd.ip);
				addr.sin_addr.s_addr = htonl(local?s->nd.local_ip:s->nd.ip);
				addr.sin_port = htons(local?s->nd.local_port:s->nd.port);
			}
			else
			{
				bool local = (getExternIp() == s->est->ip);
				addr.sin_addr.s_addr = htonl(local?s->est->local_ip:s->est->ip);
				addr.sin_port = htons(local?s->est->local_port:s->est->port);
			}
			if (addr.sin_addr.s_addr && addr.sin_port)
			{
				sendto_obfuscation(soc,(char*)buf,size,0,(sockaddr*)&addr,sizeof(sockaddr_in));
			}
		}
	}
	else // encapsulation
	{
		//if (!est) host = find_connection(s->encapsulated_id);
		//else  host = find_connection(s->est->encapsulated_id);
		//if (!host) return;
		//memcpy(buf,s->nd.id,NETWORK_ID_LENGTH_BYTES);
		//send_to_host(s,host,buf,size);
	}
}
//-----------------------------------------------------------------------
void NETWORK::send_big_buffer(CONNECTION* s, u8* buf, u32 size, u8 command)
{
	if (size > NETWORK_SIZE_DH_MAX) return;

	u8 tx_buffer[NETWORK_SIZE_UDP];
	
	u8 total_pack = (size+NETWORK_SIZE_SESSION_BUF-1)/NETWORK_SIZE_SESSION_BUF; // number of packets for sending
	for (u8 p = 0; p < total_pack; p++)
	{
		u32 chunk_size = (p==(total_pack-1)) ? (size-(total_pack-1)*NETWORK_SIZE_SESSION_BUF) : NETWORK_SIZE_SESSION_BUF; // size of data chunk
		tx_buffer[0] = command|(away?SESSION_USER_AWAY:0);
		CopyMemory(&tx_buffer[1],my_id,NETWORK_ID_LENGTH_BYTES);
		CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES],&local_ip,sizeof(u32));
		CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)],&local_port,sizeof(u16));
		tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)] = total_pack;
		tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)+1] = p;
		CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)+2],&buf[p*NETWORK_SIZE_SESSION_BUF],chunk_size);
		LogLowLevel(true,s->est->ip,s->est->port,s->nd.id,command,chunk_size);
		send_buf(s,true,tx_buffer,1+NETWORK_ID_LENGTH_BYTES+sizeof(u16)+sizeof(u32)+2+chunk_size);
		s->pack_num_sent++;
	}
}
//-----------------------------------------------------------------------
bool NETWORK::collect_big_buffer(u8* cur_buf, u32 size, ESTABLISHMENT* est)
{
	u32 total_pack = cur_buf[0];
	u32 pack_num = cur_buf[1];
	if ((pack_num >= total_pack) || (total_pack*NETWORK_SIZE_SESSION_BUF > NETWORK_SIZE_DH_MAX) ||
		((pack_num != (total_pack-1)) && (size != NETWORK_SIZE_SESSION)) || 
		((pack_num == (total_pack-1)) && (size >  NETWORK_SIZE_SESSION))) // error
	{
		est->rec_flags = 0;
		return false;
	}
	CopyMemory(&est->buff[pack_num*NETWORK_SIZE_SESSION_BUF],&cur_buf[2],size-2);
	est->rec_flags |= 1<<pack_num;
	if (pack_num == (total_pack-1)) // last packet
	{
		est->rec_size = (total_pack-1)*NETWORK_SIZE_SESSION_BUF + size-2;
	}
	for (u32 i = 0; i < total_pack; i++)
	{
		if (!(est->rec_flags & (1<<i))) return false;
	}
	return true;
}
//-----------------------------------------------------------------------
void NETWORK::send_small_buffer(CONNECTION *s, u8 cmd, const u8* buf, u32 size)
{
	u8 tx_buffer[NETWORK_SIZE_UDP];
	if (size > NETWORK_SIZE_SESSION) return;
	tx_buffer[0] = cmd|(away?SESSION_USER_AWAY:0);
	CopyMemory(&tx_buffer[1],my_id,NETWORK_ID_LENGTH_BYTES);
	CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES],&local_ip,sizeof(u32));
	CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)],&local_port,sizeof(u16));
	CopyMemory(&tx_buffer[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)],buf,size);
	if (s->est)
	{
		LogLowLevel(true,s->est->ip,s->est->port,s->nd.id,cmd,size);
		send_buf(s,true,tx_buffer,1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)+size);
		s->pack_num_sent++;
	}
	else if (s->established)
	{
		LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,cmd,size);
		send_buf(s,false,tx_buffer,1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)+size);
		s->pack_num_sent++;
	}
}
//-----------------------------------------------------------------------
u32 NETWORK::send_packet(PACKET_FOR_SEND* pack)
{
	u32 size = 0;
	u8 tx_buffer_plain[2048];
	u8 tx_buffer_coded[2048];
	CONNECTION* s;
	sockaddr_in addr = {0};
	u32 ip = pack->nd->ip;
	//if (isIPLocal(ip)&&replaceLocalIP())
	//{
	//	ip = getExternIp();
	//}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(ip);
	addr.sin_port = htons(pack->nd->port);

	if (isIdIdentical(pack->nd->id,my_id) || !session_is_initialized) return 0; // don't send data to myself or if not initialized

	EnterCriticalSection(&cs);
    if (isZeroID(pack->nd->id)) // for zero ID try to find IP:port, or just send ping
    {
		for (s = con; s; s = s->next_connection)
		{
			if ((s->nd.ip == ip) && (s->nd.port == pack->nd->port))
			{
				memcpy((void*)pack->nd->id,s->nd.id,NETWORK_ID_LENGTH_BYTES);
				break;
			}
		}
		if (!s) // can't find ID:port combination
		{
			LeaveCriticalSection(&cs);
			ping(pack->nd);
			return 0;
		}
    }
	else
	{
		s = find_connection(pack->nd->id,ip,pack->nd->port,true);
	}

	if (!s) // not enough memory for new connection
	{
		LeaveCriticalSection(&cs);
		LogMessage(module, TEXT("Not enough memory for new connection"));
		return 0;
	}

	if(s->established)
	{
		if (s->send_cnt >= NETWORK_OFFLINE_SEND_LIMIT) // so long time no any answer. Reset session
		{
			s->established = false;
			s->send_cnt = 0;
			s->ses.reset();
			ZeroMemory(s->encapsulated_id,NETWORK_ID_LENGTH_BYTES);
			if (isIdIdentical(enc_host,s->nd.id)) ZeroMemory(enc_host,NETWORK_ID_LENGTH_BYTES);
		}
		else // send encoded packet
		{
			tx_buffer_plain[0] = pack->protocol;
			CopyMemory(&tx_buffer_plain[1], pack->buffer, pack->buf_len);
			LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,SESSION_ENCODED,pack->buf_len);
			LogLowLevelEnc(true,tx_buffer_plain);
			tx_buffer_coded[0] = SESSION_ENCODED|(away?SESSION_USER_AWAY:0);
			size = s->ses.encrypt(&tx_buffer_coded[1],tx_buffer_plain,pack->buf_len+1);
			send_buf(s,false,tx_buffer_coded,1+size);
			if (!pack->dont_count_tx)
			{
				s->send_cnt++;
			}
			LeaveCriticalSection(&cs);
			return size;
		}
	}
	if (TryEnterCriticalSection(&s->est_cs))
	{
		CONNECTION* cn;
		for (cn = con; cn; cn = cn->next_connection)
		{
			if (cn != s && cn->established && ip == cn->nd.ip && pack->nd->port == cn->nd.port) // found another identity on the same IP:port, ignore first
			{
				if (s->est) delete s->est;
				s->est = NULL;
				goto dont_send_packet;
			}
		}
		if (!s->est) s->est = new ESTABLISHMENT;
		if ((s->est->state == NET_STATE_IDLE) || (s->est->state == NET_STATE_PINGING))
		{
			s->est->state = NET_STATE_PINGING;
			s->est->timeout = 50+(rand()%50);
			if (!s->est->replied)
			{
				s->est->ip = ip;
				s->est->port = pack->nd->port;
				s->est->local_ip = pack->nd->local_ip;
				s->est->local_port = pack->nd->local_port;
			}
			send_low_level_cmd(s,SESSION_PING_REQ);
			s->pack_num_sent++;
		}
		dont_send_packet:
		LeaveCriticalSection(&s->est_cs);
	}
	LeaveCriticalSection(&cs);
	return size;
}
//-----------------------------------------------------------------------
void NETWORK::forward_encapsulation(const u32* id_from, const u32* id_to, const u32* buf, u32 size)
{
	/*
	PACKET_FOR_SEND pack = {0};
	if (isIdIdentical(id_to,my_id) || !size) return;
	CONNECTION* s = find_connection(id_to);
	if (!s || !s->established) return;
	pack.buffer = new u32[size/sizeof(u32)+NETWORK_ID_LENGTH_WORDS];
	if (!pack.buffer) return;
	memcpy((u32*)&pack.buffer[0],id_from,NETWORK_ID_LENGTH_BYTES);
	memcpy((u32*)&pack.buffer[NETWORK_ID_LENGTH_WORDS],buf,size);
	pack.buf_len = size+NETWORK_ID_LENGTH_BYTES;
	pack.nd = &s->nd;
	pack.protocol = SESSION_ENC_FROM;
	send_packet(&pack);
	delete[] pack.buffer;
	*/
}
//-----------------------------------------------------------------------
s8 name_compare(const u8* buf_ref,const u8* buf_test,u32 max_size)
{
	while (max_size--)
	{
		s8 diff = *buf_ref++ - *buf_test++;
		if (diff) return diff;
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI timeoutProc(NETWORK* n)
{
	CONNECTION* s;

	while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads, 100))
	{
		EnterCriticalSection(&n->cs);
		if (!n->session_is_initialized)
		{
			LeaveCriticalSection(&n->cs);
			continue;
		}
		for (s = n->con; s; s = s->next_connection)
		{
			if (!s->established && TryEnterCriticalSection(&s->est_cs))
			{
				if (s->est)
				{
					if (s->est->timeout) s->est->timeout--;
					else
					{
						TCHAR str[400];
						TCHAR ip_port[IP_LENGTH_CCH+5];
						ip_port2str(s->est->ip,s->est->port,ip_port);
						switch(s->est->state)
						{
						case NET_STATE_PINGING:
							n->send_low_level_cmd(s,SESSION_PING_REQ);
							s->est->timeout = 100+(rand()%50);
							break;
						case NET_STATE_S_REQUESTING:
							if (s->est->send_requests)
							{
								n->send_big_buffer(s, s->est->sent_buf, s->est->sent_size, SESSION_REQUEST);
								s->est->timeout = 100+(rand()%50);
								s->est->send_requests--;
								_stprintf_s(str, TEXT("Again send session request to %s"), ip_port);
							}
							else
							{
								delete s->est;
								s->est = NULL;
								_stprintf_s(str, TEXT("Timeout for sending requests to %s"), ip_port);
							}
							break;
						case NET_STATE_S_CONFIRMING:
							if (s->est->send_requests)
							{
								n->send_big_buffer(s, s->est->conf_buf, s->est->conf_size, SESSION_CONFIRMATION);
								s->est->send_requests--;
								s->est->timeout = 100+(rand()%50);
								_stprintf_s(str, TEXT("Again send session confirmation to %s"), ip_port);
							}
							else
							{
								delete s->est;
								s->est = NULL;
								_stprintf_s(str, TEXT("Timeout for sending requests to %s"), ip_port);
							}
							break;
						default:
							delete s->est;
							s->est = NULL;
							_stprintf_s(str, TEXT("Timeout for connection from %s"), ip_port);
							break;
						}
						//LogMessage(module, str); // !!! inside critical section
					}
				}
				LeaveCriticalSection(&s->est_cs);
			}
		}
		LeaveCriticalSection(&n->cs);
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI rx_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	NETWORK* n = getNet();
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if (s->est->protocol == NETWORK_PROTOCOL_VERSION)
		{
			s->est->sent_size = s->est->rx_session.response_to_init_session_1(s->est->sent_buf,s->est->buff, getSecretData(), passw, pas_len);
			if (s->est->sent_size > 0)
			{
				n->send_big_buffer(s,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE);
				s->est->state = NET_STATE_WAIT_CONFIRMATION;
				s->est->timeout = 500+(rand()%100);
				_stprintf_s(str,TEXT("Wait confirmation because of successful session request from %s"),ip_port);
				s->est->last_size = s->est->rec_size;
				memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
				s->est->rec_size = 0;
				s->est->rec_flags = 0;
			}
			else
			{
				// WARNING: someone sent incorrect request, why?
				_stprintf_s(str,TEXT("Wrong Session request from %s"),ip_port);
				delete s->est;
				s->est = NULL;
			}
		}
		else
		{
			n->send_low_level_cmd(s,SESSION_REJECT);
			_stprintf_s(str,TEXT("Another protocol version on %s, request is ignored"),ip_port);
			delete s->est;
			s->est = NULL;
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI message_thread(void* str)
{
	MessageBox(NULL,(LPCTSTR)str,TEXT("Warning"),MB_OK|MB_ICONSTOP);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI tx_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if (s->est->protocol == NETWORK_PROTOCOL_VERSION)
		{
			s->est->conf_size = s->est->tx_session.process_response_confirm_session_1(s->est->conf_buf,s->est->buff,getSecretData(),passw,pas_len);
			if ((s->est->conf_size > 0) && s->est->tx_session.session_is_active_1(session_id))
			{
				s->userInfoSize = s->est->tx_session.get_user_info_1(s->userInfo);
				s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;

				if (isIdIdentical(s->nd.id,session_id))
				{
					s->est->state = NET_STATE_S_CONFIRMING;
					s->est->timeout = 100+(rand()%50);
					s->est->send_requests = 10;
					_stprintf_s(str,TEXT("Successful session response from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
						s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
					n->send_big_buffer(s, s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION);
					s->est->last_size = s->est->rec_size;
					memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
					s->est->rec_size = 0;
					s->est->rec_flags = 0;
				}
				else // alarm!!!
				{
					s->established = false;
					_stprintf_s(str,TEXT("USER ID IS NOT THE SAME AS KEY FINGERPTING from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
						s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
					delete s->est;
					s->est = NULL;
					CloseHandle(CreateThread(NULL,0,message_thread,str,0,NULL));
					Sleep(50);
				}
			}
			else
			{
				// WARNING: Someone sent incorrect response
				_stprintf_s(str,TEXT("Wrong Session response from %s"),ip_port);
				delete s->est;
				s->est = NULL;
			}
		}
		else
		{
			_stprintf_s(str,TEXT("Another protocol version on %s"),ip_port);
			delete s->est;
			s->est = NULL;
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI conf_signature(CONNECTION* s)
{
	EnterCriticalSection(&s->est_cs);
	TCHAR str[400];
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR ip_port[IP_LENGTH_CCH+5];
	if (s->est)
	{
		u32 size = s->est->rx_session.check_confirmation_1(s->est->buff);
		ip_port2str(s->est->ip,s->est->port,ip_port);
		if ((size > 0) && s->est->rx_session.session_is_active_1(session_id)) // normal response
		{
			s->userInfoSize = s->est->rx_session.get_user_info_1(s->userInfo);
			s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;
			if (isIdIdentical(s->nd.id,session_id))
			{
				s->est->rx_session.start_initialized_session(&s->ses);
				s->established = true; // here we start the new established session
				s->nd.ip = s->est->ip;
				s->nd.port = s->est->port;
				s->nd.local_ip = s->est->local_ip;
				s->nd.local_port = s->est->local_port;
				memcpy(s->encapsulated_id,s->est->encapsulated_id,NETWORK_ID_LENGTH_BYTES);
				s->send_cnt = 0;
				if (isZeroID(n->enc_host)) memcpy(n->enc_host,s->nd.id,sizeof(n->enc_host));
				n->send_small_buffer(s,SESSION_CONFIRMED|(n->away?SESSION_USER_AWAY:0),NULL,0);

				_stprintf_s(str,TEXT("Successful session confirmation from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
					s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
				
			}
			else // alarm!!!
			{
				s->established = false;
				_stprintf_s(str,TEXT("USER ID IS NOT THE SAME AS KEY FINGERPTING from %s, user name: %s, version %d, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),ip_port,
					s->userInfo,s->est->version,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
				CloseHandle(CreateThread(NULL,0,message_thread,str,0,NULL));
				Sleep(50);
			}
		}
		else
		{
			_stprintf_s(str,TEXT("Wrong Session confirmation from %s"),ip_port);
		}
		delete s->est;
		s->est = NULL;
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	}
	else LeaveCriticalSection(&s->est_cs);
	return 0;
}
//---------------------------------------------------------------------------------------------------------------------------------------------
void NETWORK::send_low_level_cmd(CONNECTION* s, u8 cmd)
{
	u8 buf[3];
	buf[0] = NETWORK_PROTOCOL_VERSION;
	*(u16*)&buf[1] = BUILD_NUM;
	send_small_buffer(s,cmd|(away?SESSION_USER_AWAY:0),buf,sizeof(buf));
}
//---------------------------------------------------------------------------------------------------------------------------------------------
void NETWORK::ProcessRx(u8* buffer,int pack_len,u32 ip,u16 port,const u32* host_id)
{
	TCHAR str[400];
	TCHAR ip_port[IP_LENGTH_CCH+5];
	u8 buffer_plain[2*NETWORK_SIZE_UDP];
	CONNECTION* s;
	KADEMLIA* kad = getKad();
	u8 cmd = buffer[0]&~SESSION_USER_AWAY;
	u8 away = buffer[0]&SESSION_USER_AWAY;
	buffer++;
	pack_len--;

//	if (isIPLocal(ip)&&replaceLocalIP())
//	{
//		ip = getExternIp();
//	}

    EnterCriticalSection(&cs);
	if (!session_is_initialized)
	{
		LeaveCriticalSection(&cs);
		return;
	}
	if (cmd < SESSION_ENCODED) // non-coded packet. ID is just after command
	{
		if ((pack_len<NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16))||isIdIdentical((u32*)buffer,my_id)||isZeroID((u32*)buffer)) // check packet validity
		{
			LeaveCriticalSection(&cs);
			return;
		}
	    s = find_connection((u32*)buffer,ip,port,true);
		if (TryEnterCriticalSection(&s->est_cs))
		{
			if (!s->est) s->est = new ESTABLISHMENT;
			s->est->ip = ip;
			s->est->port = port;
			s->est->local_ip = *(u32*)&buffer[NETWORK_ID_LENGTH_BYTES];
			s->est->local_port = *(u16*)&buffer[NETWORK_ID_LENGTH_BYTES+sizeof(u32)];
			s->est->replied = true;
			LeaveCriticalSection(&s->est_cs);
		}
		buffer+=NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16);
		pack_len-=NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16);
	}
	else // encoded packet, find connection by IP:port
	{
		s = find_connection(ip,port);
	}
	if (!s) // can't find such connection or memory problems
	{
		LeaveCriticalSection(&cs);
		return;
	}
	LogLowLevel(false,ip,port,s->nd.id,cmd,pack_len);
	s->away = (away==SESSION_USER_AWAY);
	ip_port2str(ip,port,ip_port);
	str[0] = 0;
	if (kad) kad->updateNode(&s->nd,true,getUserList());
    switch(cmd)
    {
    case SESSION_PING_REQ: //---------------------------------------------------------------------------------------------------------------------------------------------
		if (pack_len < 3) break; // check packet size
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (!s->est) s->est = new ESTABLISHMENT;
			s->est->protocol = buffer[0];
			s->est->version = *(u16*)&buffer[1];
			if (host_id) memcpy(s->est->encapsulated_id,host_id,NETWORK_ID_LENGTH_BYTES);
			if (s->est->protocol > NETWORK_PROTOCOL_VERSION)
			{
				send_low_level_cmd(s,SESSION_REJECT);
				_stprintf_s(str,TEXT("Another protocol version %d on %s"),s->est->protocol,ip_port);
				delete s->est;
				s->est = NULL;
			}
			else
			{
				switch(s->est->state)
				{
				case NET_STATE_IDLE:
				case NET_STATE_P_REQUESTED:
				case NET_STATE_REJECTED:
					s->est->state = NET_STATE_P_REQUESTED;
					s->est->timeout = 50+(rand()%50);
					send_low_level_cmd(s,SESSION_PING_RESP);
					break;
				case NET_STATE_PINGING:
				case NET_STATE_S_REQUESTING:
				case NET_STATE_S_CONFIRMING:
					if (s->nd.id[0] > my_id[0]) // both are trying to connect.
					{
						s->est->state = NET_STATE_P_REQUESTED;
						s->est->timeout = 50+(rand()%50);
						send_low_level_cmd(s,SESSION_PING_RESP);
						_stprintf_s(str,TEXT("Conflicted ping request from %s, switched to passive mode"),ip_port);
					}
					else
					{
						send_low_level_cmd(s,SESSION_REJECT);
						_stprintf_s(str,TEXT("Conflicted ping request from %s, request rejected"),ip_port);
					}
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
        break;
    case SESSION_PING_RESP: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->established) // check that nobody else has the same identity
			{
				CONNECTION* ptr,*prev_s;
				for (prev_s = ptr = con; ptr;)
				{
					if (ptr != s && (((ptr->nd.ip == ip) && (ptr->nd.port == port)) || 
						(ptr->est && (ptr->est->ip == ip) && (ptr->est->port == port))))
					{
						// delete connection
						if (ptr == con)
						{
							con = con->next_connection;
						}
						else
						{
							prev_s->next_connection = ptr->next_connection;
						}
						getUserList()->delete_node(&ptr->nd);
						if (kad) kad->deleteNode(&ptr->nd);
						delete ptr;
						ptr = prev_s->next_connection;
						break;
					}
					else
					{
						prev_s = ptr;
						ptr = ptr->next_connection;
					}
				}
			}
			if (s->established && !s->est) // If session is established and ping responce is not expected: ignore it.
			{
				LeaveCriticalSection(&s->est_cs);
				break;
			}
			if (!s->est) s->est = new ESTABLISHMENT;
			switch(s->est->state)
			{
			case NET_STATE_IDLE:
			case NET_STATE_PINGING:
				s->est->protocol = buffer[0];
				s->est->version = *(u16*)&buffer[1];
				if (host_id) memcpy(s->est->encapsulated_id,host_id,NETWORK_ID_LENGTH_BYTES);
				if (s->est->protocol <= NETWORK_PROTOCOL_VERSION)
				{
					s->est->state = NET_STATE_S_REQUESTING;
					s->est->timeout = 100+(rand()%50);
					s->est->send_requests = 5;
					s->est->tx_session = cmn_tx_session;
					s->est->sent_size = sending_size;
					CopyMemory(s->est->sent_buf,sending_buff,sending_size);
					send_big_buffer(s,s->est->sent_buf,s->est->sent_size,SESSION_REQUEST);
					_stprintf_s(str,TEXT("Send session request to %s"),ip_port);
				}
				else
				{
					_stprintf_s(str,TEXT("Another protocol version %d on %s"),s->est->protocol,ip_port);
					delete s->est;
					s->est = NULL;
				}
				break;
			}
			LeaveCriticalSection(&s->est_cs);
		}
        break;
    case SESSION_REQUEST: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				switch(s->est->state)
				{
				case NET_STATE_PINGING:
				case NET_STATE_S_REQUESTING:
				case NET_STATE_S_CONFIRMING:
					if (s->nd.id[0] > my_id[0]) // both are trying to connect.
					{
						s->est->state = NET_STATE_REJECTED;
						s->est->timeout = 300+(rand()%100);
						_stprintf_s(str,TEXT("Conflicted session requrest from %s, switched to passive mode"),ip_port);
					}
					else
					{
						send_low_level_cmd(s,SESSION_REJECT);
						_stprintf_s(str,TEXT("Conflicted session requrest from %s, request rejected"),ip_port);
					}
					break;
				case NET_STATE_P_REQUESTED:
				case NET_STATE_REJECTED:
					s->est->state = NET_STATE_S_REQUESTED;
					s->est->timeout = 300+(rand()%100);
				case NET_STATE_S_REQUESTED:
					if (collect_big_buffer(buffer,pack_len,s->est))
					{
						bool same_request = false;
						s->est->timeout=500+(rand()%100);
						if (s->est->last_size && s->est->sent_size && (s->est->last_size == s->est->rec_size))
						{
							same_request = true;
							for (u32 i = 0; i < s->est->rec_size/4; i++)
								if (s->est->buff[i] != s->est->last_buf[i])
								{
									same_request = false;
									break;
								}
						}
						if (same_request)
						{
							send_big_buffer(s,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE);
							LeaveCriticalSection(&s->est_cs);
							LeaveCriticalSection(&cs);
							_stprintf_s(str,TEXT("Same session request from %s"),ip_port);
							s->est->rec_size = 0;
							s->est->rec_flags = 0;
							break;
						}
						else
						{
							CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)rx_signature,s,0,NULL));
						}
					}
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
		else
		{
			send_small_buffer(s,SESSION_BUSY|(away?SESSION_USER_AWAY:0),NULL,0);
		}
        break;
    case SESSION_RESPONSE: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				switch(s->est->state)
				{
				case NET_STATE_S_REQUESTING:
					if (collect_big_buffer(buffer,pack_len,s->est))
					{
						CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)tx_signature,s,0,NULL));
					}
					break;
				case NET_STATE_S_CONFIRMING:
					send_big_buffer(s,s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION);
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
        break;
    case SESSION_CONFIRMATION: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				switch(s->est->state)
				{
				case NET_STATE_WAIT_CONFIRMATION:
					if (collect_big_buffer(buffer,pack_len,s->est))
					{
						CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)conf_signature,s,0,NULL));
					}
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
		else
		{
			send_small_buffer(s,SESSION_BUSY|(away?SESSION_USER_AWAY:0),NULL,0);
		}
        break;
	case SESSION_BUSY: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				switch(s->est->state)
				{
				case NET_STATE_S_REQUESTING:
				case NET_STATE_S_CONFIRMING:
					s->est->timeout = 300+(rand()%100);
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
		break;
    case SESSION_REJECT: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				s->est->protocol = buffer[0];
				s->est->version = *(u16*)&buffer[1];
				s->est->state = NET_STATE_REJECTED;
				s->est->timeout = 300+(rand()%100);
				_stprintf_s(str,TEXT("Rejected connection to %s"),ip_port);
			}
			LeaveCriticalSection(&s->est_cs);
		}
		break;
    case SESSION_CONFIRMED: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est)
			{
				switch(s->est->state)
				{
				case NET_STATE_S_CONFIRMING:
					s->est->tx_session.start_initialized_session(&s->ses);
					s->established = true; // here we start the new established session
					s->nd.ip = s->est->ip;
					s->nd.port = s->est->port;
					s->nd.local_ip = s->est->local_ip;
					s->nd.local_port = s->est->local_port;
					memcpy(s->encapsulated_id,s->est->encapsulated_id,NETWORK_ID_LENGTH_BYTES);
					if (isZeroID(enc_host)) memcpy(enc_host,s->nd.id,sizeof(enc_host));
					s->send_cnt = 0;
					_stprintf_s(str,TEXT("Session confirmed by %s"),ip_port);
					break;
				}
			}
			LeaveCriticalSection(&s->est_cs);
		}
		break;
    case SESSION_TERMINATE: //---------------------------------------------------------------------------------------------------------------------------------------------
		s->pack_num_est++;
		if(s->established && TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est) delete s->est;
			s->est = NULL;
			s->established = false;
			s->ses.reset();
			ZeroMemory(s->encapsulated_id,NETWORK_ID_LENGTH_BYTES);
			if (isIdIdentical(enc_host,s->nd.id)) ZeroMemory(enc_host,NETWORK_ID_LENGTH_BYTES);
			LeaveCriticalSection(&s->est_cs);
			_stprintf_s(str,TEXT("Session rerminated by %s"),ip_port);
		}
		break;
    case SESSION_ENCODED: //---------------------------------------------------------------------------------------------------------------------------------------------
	case SESSION_NAME_REQ:
	case SESSION_NAME_RESP:
	case SESSION_ENC_FOR:
	case SESSION_ENC_FROM:
		if (!s->established && TryEnterCriticalSection(&s->est_cs))
		{
			if (s->est && s->est->state == NET_STATE_S_CONFIRMING) // session established normaly, don't need data anymore
			{
				s->est->tx_session.start_initialized_session(&s->ses);
				s->established = true; // here we start the new established session
				s->nd.ip = s->est->ip;
				s->nd.port = s->est->port;
				s->nd.local_ip = s->est->local_ip;
				s->nd.local_port = s->est->local_port;
				memcpy(s->encapsulated_id,s->est->encapsulated_id,NETWORK_ID_LENGTH_BYTES);
				if (isZeroID(enc_host)) memcpy(enc_host,s->nd.id,sizeof(enc_host));
				s->send_cnt = 0;
				_stprintf_s(str,TEXT("Session confirmed by %s"),ip_port);
			}
			LeaveCriticalSection(&s->est_cs);
		}
        if (s->established)
        {
            u32 decr_size = s->ses.decrypt(buffer_plain,buffer,pack_len);
            if (decr_size > sizeof(u8))
            {
				s->send_cnt = 0;
				LogLowLevelEnc(false,buffer_plain);
				if (TryEnterCriticalSection(&s->est_cs))
				{
					if (s->est) // session established normaly, don't need data anymore
					{
						delete s->est;
						s->est = NULL;
					}
					LeaveCriticalSection(&s->est_cs);
				}
				s->pack_num_sec++;
				switch (cmd)
				{
				case SESSION_ENCODED:
				{
					NET_PACKET np;
					np.buffer = buffer_plain;
					np.buf_len = decr_size;
					np.nd = &s->nd;
					LeaveCriticalSection(&cs);
					if (WAIT_TIMEOUT == WaitForSingleObject(eStopThreads,0))
					{
						parsePacket(&np);
					}
					return;
				}
				case SESSION_NAME_REQ:
					if (decr_size >= *(u32*)buffer_plain && s->established)
					{
						buffer_plain[4+*(u32*)buffer_plain] = 0;
						buffer_plain[4+*(u32*)buffer_plain+1] = 0;
						u32 partID = (u32)_tcstoll((LPTSTR)&buffer_plain[4],NULL,16);
						CONNECTION* co;
						for (co = con; co; co = co->next_connection)
						{
							bool found = !name_compare(&buffer_plain[4],co->userInfo,min(co->userInfoSize,*(u32*)buffer_plain));
							if (!found && partID)
							{
								for (int i = 0; i < NETWORK_ID_LENGTH_WORDS; i++)
								{
									if (partID == s->nd.id[i])
									{
										found = true;
										break;
									}
								}
							}
							if (found)
							{
								u32 size = s->ses.encrypt(buffer_plain,(u8*)co->nd.id,NETWORK_ID_LENGTH_BYTES);
								send_small_buffer(s,SESSION_NAME_RESP,buffer_plain,size);
								break;
							}
						}
					}
					break;
				case SESSION_NAME_RESP:
					LeaveCriticalSection(&cs);
					find_user_responces++;
					{
						u32 total = 0;
						CONNECTION* co;
						for (co = con; co; co = co->next_connection)
						{
							if (co->established) total++;
						}
						if (total>0)
							PostMessage(mainWnd,WM_USER_STEP,(WPARAM)find_user_responces*100/total,0);
					}
					if (decr_size >= NETWORK_ID_LENGTH_BYTES)
					{
						u32* id = new u32[NETWORK_ID_LENGTH_WORDS];
						CopyMemory(id,buffer_plain,NETWORK_ID_LENGTH_BYTES);
						PostMessage(mainWnd,WM_USER_FOUND,0,(LPARAM)id);
					}
					return;
/*
				case SESSION_ENC_FOR:
					if (decr_size > sizeof(u32)*(1+NETWORK_ID_LENGTH_WORDS))
					{
						forward_encapsulation(nd->id,buffer_plain,&buffer_plain[NETWORK_ID_LENGTH_WORDS],decr_size-sizeof(u32)*NETWORK_ID_LENGTH_WORDS);
					}
					break;
				case SESSION_ENC_FROM:
					if (decr_size >= sizeof(u32)*(1+NETWORK_ID_LENGTH_WORDS))
					{
						NODE nd_from = {0};
						memcpy(nd_from.id,buffer_plain,sizeof(u32)*NETWORK_ID_LENGTH_WORDS);
						LeaveCriticalSection(&cs);
						ProcessRx(&buffer_plain[NETWORK_ID_LENGTH_WORDS],decr_size-sizeof(u32)*NETWORK_ID_LENGTH_WORDS,&nd_from,nd->id);
						return;
					}
					break;
				*/
				}
            }
            else
			{
				s->pack_num_drop++;
			}
        }
        break;
    }
	LeaveCriticalSection(&cs);
	if (str[0]) LogMessage(module,str);
}
//-----------------------------------------------------------------------
#include "pjlib.h"
DWORD WINAPI receiverProc(NETWORK* n)
{
	int pack_len;
	sockaddr_in addr = {0};
	u8 buffer[2*NETWORK_SIZE_UDP];
	pj_thread_t* pj_thr;
	pj_thread_desc pj_desc = {1};
	pj_thread_register("rec",pj_desc,&pj_thr);

    while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads,0))
	{
		int addr_size = sizeof(addr);
		memset(&addr,0,sizeof(addr));
		addr.sin_family = AF_INET;
		//pack_len = recvfrom(n->soc,(char*)buffer_coded,sizeof(buffer_coded),0,(sockaddr*)&addr,&addr_size);
		pack_len = recvfrom_obfuscation(n->soc,(char*)buffer,sizeof(buffer),0,(sockaddr*)&addr,&addr_size);
		if (pack_len == SOCKET_ERROR) // error parsing
        {
			int err;
			if ((err = WSAGetLastError()) == WSAETIMEDOUT) continue;
			closesocket(n->soc);
			// open new socket
			n->soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
			if (n->soc == INVALID_SOCKET)
			{
				return 0;
			}
			if (SOCKET_ERROR == bind(n->soc,(sockaddr*)&n->my_addr,sizeof(n->my_addr)))
			{
				return 0;
			}
			// disable  new behavior using
			// IOCTL: SIO_UDP_CONNRESET
			DWORD dwBytesReturned = 0;
			BOOL bNewBehavior = FALSE;
			WSAIoctl(n->soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
			// set receiver timeout in ms
			DWORD timeout = 100;
			setsockopt(n->soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
			continue;
        }
        if (pack_len > 0) n->ProcessRx(buffer,pack_len,ntohl(addr.sin_addr.S_un.S_addr),ntohs(addr.sin_port),NULL);
	}
	return 0;
}
//---------------------------------------------------
CONNECTION* NETWORK::get_next_node(CONNECTION* cur)
{
    CONNECTION* curc = cur;
    if (cur)
    {
        return cur->next_connection;
    }
    return cur;
}
//---------------------------------------------------
void NETWORK::get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent, u32* missed, u32* late, u32* invalid)
{
	*est = 0;
	*sec = 0;
	*drop = 0;
	*sent = 0;
	*missed = 0;
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd->id);
	if (s)
	{
		*est = s->pack_num_est;
		*sec = s->pack_num_sec;
		*drop = s->pack_num_drop;
		*sent = s->pack_num_sent;
		*missed = (u32)s->ses.dropped_rx_bytes();
		*late = (u32)s->ses.late_packets();
		*invalid = (u32)s->ses.invalid_packets();
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::get_user_info(NODE* nd, LPTSTR info)
{
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd->id);
	if (s)
	{
		memcpy(info,s->userInfo,s->userInfoSize);
		info[s->userInfoSize/sizeof(TCHAR)]=0;
	}
	else
	{
		cp_strcchcpy(info,user_info_max_length_bytes/sizeof(tchar_t),_T("<unknown>"));
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
bool NETWORK::findUserByName(u8* userInfo, u16 size, u32* id)
{
	CONNECTION *s;
	u32 tx_buffer_plain[200];
	u8 tx_buffer_coded[NETWORK_SIZE_ENCODED];
	sockaddr_in addr = {0};
	u32 partID = (u32)_tcstoll((LPTSTR)userInfo,NULL,16);
	//
	if (size > user_info_max_length_bytes) return false;
    EnterCriticalSection(&cs);
	find_user_responces = 0;
	// try to find locally
	for (s = con; s; s = s->next_connection)
	{
		bool found = !name_compare(userInfo,s->userInfo,min(s->userInfoSize,size));
		if (!found && partID)
		{
			for (int i = 0; i < NETWORK_ID_LENGTH_WORDS; i++)
			{
				if (partID == s->nd.id[i])
				{
					found = true;
					break;
				}
			}
		}
		if (found)
		{
			u32* id_msg = new u32[NETWORK_ID_LENGTH_WORDS];
			memcpy(id,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH_WORDS);
			memcpy(id_msg,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH_WORDS);
			LeaveCriticalSection(&cs);
			PostMessage(mainWnd,WM_USER_FOUND,0,(LPARAM)id_msg);
			return true;
		}
	}
	// try to find externally
	for (s = con; s; s = s->next_connection)
	{
		if(s->established)
		{
			addr.sin_family = AF_INET;
			addr.sin_addr.S_un.S_addr = htonl(s->nd.ip);
			addr.sin_port = htons(s->nd.port);
			tx_buffer_plain[0] = size;
			CopyMemory(&tx_buffer_plain[1], userInfo, size);
			size = s->ses.encrypt((u8*)tx_buffer_coded,(u8*)tx_buffer_plain,((size+3)/4)*4+sizeof(u32));
			send_small_buffer(s,SESSION_NAME_REQ,tx_buffer_coded,size);
		}	
	}
	LeaveCriticalSection(&cs);
	return false;
}
//---------------------------------------------------
int NETWORK::get_user_status(const u32* id)
{
	CONNECTION *s;
	int res = USER_STATUS_OFFLINE;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			if (s->established) res = s->away?USER_STATUS_AWAY:USER_STATUS_ONLINE;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
bool NETWORK::connected_to_user(const u32* id)
{
	CONNECTION *s;
	//
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			if (s->established) return true;
			break;
		}
	}
	return false;
}
//---------------------------------------------------
u16* NETWORK::get_session_hash(const u32* id)
{
	CONNECTION *s;
	u16* ptr = NULL;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			ptr = (u16*)s->ses.session_fingerprint(); // 256-bit value
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return ptr;
}
//---------------------------------------------------
bool NETWORK::findNode(const u32 *id, NODE *nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			*nd = s->nd;
			res = true;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
void NETWORK::get_ip(NODE* nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(nd->id,s->nd.id))
		{
			nd->ip = s->nd.ip;
			nd->port = s->nd.port;
			break;
		}
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::set_ip(NODE* nd)
{
	CONNECTION *s;
	bool new_node = true;
	u32 old_ip = 0;
	u16 old_port = 0;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(nd->id,s->nd.id))
		{
			new_node = false;
			if (!s->established && nd->ip && nd->port)
			{
				old_ip = s->nd.ip;
				old_port = s->nd.port;
				s->nd.ip = nd->ip;
				s->nd.port = nd->port;
			}
			break;
		}
	}
	LeaveCriticalSection(&cs);
	if (new_node) // new node, nevere seen before
	{
		u32 msg = 1;
		PACKET_FOR_SEND pack = {0};
		pack.buffer = (u8*)&msg;
		pack.buf_len = sizeof(u32);
		pack.nd = nd;
		pack.protocol = NETWORK_PROT_KADEMLIA;
		send_packet(&pack);
	}
	else if ((old_ip!=nd->ip)||(old_port!=nd->port)) // IP is updated
	{
		TCHAR log_msg[1000];
		getUserList()->update_node(nd);
		_stprintf_s(log_msg,TEXT("Updated IP and port for %08X %08X %08X %08X %08X %08X %08X %08X from %3d.%3d.%3d.%3d:%5d to %3d.%3d.%3d.%3d:%5d\n"),nd->id[0],nd->id[1],nd->id[2],nd->id[3],nd->id[4],nd->id[5],nd->id[6],nd->id[7],
			FIRST_IPADDRESS(old_ip),SECOND_IPADDRESS(old_ip),THIRD_IPADDRESS(old_ip),FOURTH_IPADDRESS(old_ip),old_port,
			FIRST_IPADDRESS(nd->ip),SECOND_IPADDRESS(nd->ip),THIRD_IPADDRESS(nd->ip),FOURTH_IPADDRESS(nd->ip),nd->port);
		LogString(log_msg);
	}
}
//---------------------------------------------------
void NETWORK::reset_sessions(void)
{
	CONNECTION *s;
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
    EnterCriticalSection(&cs);
	session_is_initialized = false;
    LeaveCriticalSection(&cs);
	if (pas_len)
	{
		sending_size = cmn_tx_session.init_session_1(sending_buff);
	}
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		s->established = 0;
		s->ses.reset();
		EnterCriticalSection(&s->est_cs);
		if (s->est) delete s->est;
		s->est = NULL;
		LeaveCriticalSection(&s->est_cs);
	}
	session_is_initialized = true;
	LeaveCriticalSection(&cs);
}
//-----------------------------------------------------------------------
bool NETWORK::findNodeByIpPort(NODE* nd)
{
	bool res = false;
    EnterCriticalSection(&cs);
	for (CONNECTION *s = con; s; s = s->next_connection)
	{
		if ((s->nd.ip == nd->ip) && (s->nd.port == nd->port))
		{
			memcpy(nd->id,s->nd.id,NETWORK_ID_LENGTH_BYTES);
			res = true;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//-----------------------------------------------------------------------
void NETWORK::ping(const NODE* nd)
{
	BOOL send_directly = true;
    EnterCriticalSection(&cs);
	CONNECTION *s = find_connection(nd->id);
	if (s && TryEnterCriticalSection(&s->est_cs))
	{
		if (s->est && s->est->state != NET_STATE_IDLE) send_directly = false;
		else
		{
			CONNECTION* cn;
			for (cn = con; cn; cn = cn->next_connection)
			{
				if (cn != s && s->nd.ip == cn->nd.ip && s->nd.port == cn->nd.port && cn->established) // found another identity on the same IP:port, ignore first
				{
					if (s->est) delete s->est;
					s->est = NULL;
					send_directly = false;
					break;
				}
			}
		}
		LeaveCriticalSection(&s->est_cs);
	}
	LeaveCriticalSection(&cs);
	if (send_directly)
	{
		sockaddr_in addr = {0};
		addr.sin_family = AF_INET;
		addr.sin_addr.S_un.S_addr = htonl(nd->ip);
		addr.sin_port = htons(nd->port);
		u8 buf[NETWORK_ID_LENGTH_BYTES+4+6];
		buf[0] = SESSION_PING_REQ|(away?SESSION_USER_AWAY:0);
		memcpy(&buf[1],my_id,NETWORK_ID_LENGTH_BYTES);
		memcpy(&buf[1+NETWORK_ID_LENGTH_BYTES],&local_ip,sizeof(u32));
		memcpy(&buf[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)],&local_port,sizeof(u16));
		buf[1+NETWORK_ID_LENGTH_BYTES+sizeof(u32)+sizeof(u16)] = NETWORK_PROTOCOL_VERSION;
		*(u16*)&buf[NETWORK_ID_LENGTH_BYTES+2+sizeof(u32)+sizeof(u16)] = BUILD_NUM;
		LogLowLevel(true,nd->ip,nd->port,nd->id,SESSION_PING_REQ,3);
		sendto_obfuscation(soc,(char*)buf,sizeof(buf),0,(sockaddr*)&addr,sizeof(sockaddr_in));
	}
}
//---------------------------------------------------
