#include <windows.h>
#include <commctrl.h>
#include <tchar.h>
#include <Shlwapi.h>
#include <Windns.h>
#include <shlobj.h>
#include <Windowsx.h>
#include <math.h>

#include "resource.h"
#include "settings.h"
#include "upnp.h"
#include "log.h"
#include "kademlia.h"
#include "main.h"
#include "password.h"
#include "chat.h"
#include "..\audio\waveio.h"
#include "..\audio\wav_def.h"
#include "build.h"
#include "video.h"
//----------------------------------------------------------------------------
void SelectFile(HWND hWnd, UINT idc)
{
	TCHAR FileName[MAX_PATH];
	OPENFILENAME File = {0};
	GetDlgItemText(hWnd,idc,FileName,MAX_PATH);
	File.lStructSize = sizeof(OPENFILENAME);
	File.hwndOwner = hWnd;
	File.lpstrFile = FileName;
	File.nMaxFile = MAX_PATH;
	File.lpstrFilter = TEXT("ZAS files (*.zas)\0*.zas\0*.*\0*.*\0");
	File.nFilterIndex = 1;
	File.Flags = OFN_PATHMUSTEXIST;
   	if (!GetOpenFileName(&File))
   		return;
	SetDlgItemText(hWnd,idc,FileName);
}
//----------------------------------------------------------------------------
void SelectFolder(HWND hWnd, UINT idc)
{
	TCHAR path[MAX_PATH];
	BROWSEINFO bi;
	PIDLIST_ABSOLUTE pl;
	bi.hwndOwner = hWnd;
	bi.pidlRoot = NULL;
	bi.pszDisplayName = NULL;
	bi.lpszTitle = TEXT("Select default folder for file transfer");
	bi.ulFlags = BIF_NEWDIALOGSTYLE;
	bi.lpfn = NULL;
	bi.lParam = NULL;
	bi.iImage = 0;
	pl = SHBrowseForFolder(&bi);
	if (!pl || !SHGetPathFromIDList(pl, path))
		return;
	SetDlgItemText(hWnd, idc, path);
}
//----------------------------------------------------------------------------
static void SetFileInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd,TEXT("set"),set);
	SetDlgItemText(hWnd,IDC_SET_KEYFILE,set->keyFileName);
	SetDlgItemText(hWnd,IDC_SET_NETFILE,set->userFileName);
	SetDlgItemText(hWnd,IDC_SET_LIKESFILE,set->likesFileName);
	SetDlgItemText(hWnd,IDC_SET_LOGFILE,set->logFileName);
	SetDlgItemText(hWnd,IDC_SET_DOWNLOADFILE,set->downloadsFileName);
	SetDlgItemText(hWnd,IDC_SET_SEEDFILE,set->seedFileName);
	SetDlgItemText(hWnd, IDC_SET_FTPATH, set->ftPath);
	CheckDlgButton(hWnd, IDC_LOG_ENABLE, set->enableLogFile ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_SET_FTAUTO, set->reciveFiles ? BST_CHECKED : BST_UNCHECKED);
	SetDlgItemInt(hWnd,IDC_LOG_SIZE,set->log_size,FALSE);
}
//----------------------------------------------------------------------------
static void SetFileDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd,TEXT("set"));
	GetDlgItemText(hWnd,IDC_SET_KEYFILE,set->keyFileName,MAX_PATH);
	GetDlgItemText(hWnd,IDC_SET_NETFILE,set->userFileName,MAX_PATH);
	GetDlgItemText(hWnd,IDC_SET_LIKESFILE,set->likesFileName,MAX_PATH);
	GetDlgItemText(hWnd,IDC_SET_LOGFILE,set->logFileName,MAX_PATH);
	GetDlgItemText(hWnd,IDC_SET_DOWNLOADFILE,set->downloadsFileName,MAX_PATH);
	GetDlgItemText(hWnd,IDC_SET_SEEDFILE,set->seedFileName,MAX_PATH);
	GetDlgItemText(hWnd, IDC_SET_FTPATH, set->ftPath, MAX_PATH);
	set->enableLogFile = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_LOG_ENABLE));
	set->reciveFiles = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_SET_FTAUTO));
	set->log_size = GetDlgItemInt(hWnd,IDC_LOG_SIZE,NULL,FALSE);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetFileProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetFileInit(hWnd,(settings_t*)((PROPSHEETPAGE*)lParam)->lParam);
		return TRUE;
		//
    case WM_DESTROY:
		SetFileDestroy(hWnd);
		return 0;
		//
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_SET_KEYFILESEL:
		case IDC_SET_NETFILESEL:
		case IDC_SET_LIKESFILESEL:
		case IDC_SET_LOGFILESEL:
		case IDC_SET_DOWNLOADFILESEL:
		case IDC_SET_SEEDFILESEL:
			SelectFile(hWnd, LOWORD(wParam) - 1);
			return TRUE;
		case IDC_SET_FTPATHSEL:
			SelectFolder(hWnd, LOWORD(wParam) - 1);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------/----------------------------------------------------------------------------
static void SetNetworkInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd,TEXT("set"),set);
	SetDlgItemInt(hWnd,IDC_SET_LOCPORT,set->localPort,FALSE);
	SetDlgItemInt(hWnd,IDC_SET_REMPORT,set->remotePort,FALSE);
	CheckDlgButton(hWnd,IDC_UPNP_ENABLE,set->enableUPnP?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_UPNP_ON_EXIT,set->removeUPnPonExit?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_UPNP_NEXT_PORT,set->nextAvailablePort?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_UPNP_RANDOM,set->randomPort?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_OBFUSCATION,set->obfuscationEnabled?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_ENCAPSULATION,set->encapsulation?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_REPLACEIP,set->replaceLocalIP?BST_CHECKED:BST_UNCHECKED);
}
//----------------------------------------------------------------------------
static void SetNetworkDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd,TEXT("set"));
	set->localPort = GetDlgItemInt(hWnd,IDC_SET_LOCPORT,NULL,FALSE);
	set->remotePort = GetDlgItemInt(hWnd,IDC_SET_REMPORT,NULL,FALSE);
	set->enableUPnP = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_UPNP_ENABLE));
	set->removeUPnPonExit = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_UPNP_ON_EXIT));
	set->nextAvailablePort = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_UPNP_NEXT_PORT));
	set->randomPort = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_UPNP_RANDOM));
	set->obfuscationEnabled = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_OBFUSCATION));
	set->encapsulation = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_ENCAPSULATION));
	set->replaceLocalIP= (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_REPLACEIP));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetNetworkProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetNetworkInit(hWnd,(settings_t*)((PROPSHEETPAGE*)lParam)->lParam);
		return TRUE;
		//
    case WM_DESTROY:
		SetNetworkDestroy(hWnd);
		return 0;
		//
	}
	return FALSE;
}
//----------------------------------------------------------------------------/----------------------------------------------------------------------------
static void SetPrivacyInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd, TEXT("set"), set);
	CheckDlgButton(hWnd, IDC_SET_ALLOW_ANYONE, set->allowAnyone ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_SET_SHOW_REQUEST, set->allowConnection ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_SET_ACCEPT_FT, set->acceptFT ? BST_CHECKED : BST_UNCHECKED);
}
//----------------------------------------------------------------------------
static void SetPrivacyDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd, TEXT("set"));
	set->allowAnyone = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_SET_ALLOW_ANYONE));
	set->allowConnection = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_SET_SHOW_REQUEST));
	set->acceptFT = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_SET_ACCEPT_FT));
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetPrivacyProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: SetPrivacyInit(hWnd, (settings_t*)((PROPSHEETPAGE*)lParam)->lParam); return TRUE;
	case WM_DESTROY: SetPrivacyDestroy(hWnd); return 0;
	}
	return FALSE;
}
//----------------------------------------------------------------------------/----------------------------------------------------------------------------
static void SetUserInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd, TEXT("set"), set);
	SetDlgItemInt(hWnd,IDC_SET_REMEMBER,set->passwordTime,FALSE);
	if (set->rememberPassword) EnableWindow(GetDlgItem(hWnd,IDC_SET_REMEMBER),FALSE);
	CheckDlgButton(hWnd, IDC_STARTWIN, set->autostart ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_PREVENTSLEEP, set->preventsleep ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_FWPORTS, set->openfirewall ? BST_CHECKED : BST_UNCHECKED);
	SetDlgItemInt(hWnd,IDC_SET_AWAYTIME,set->away_timeout,FALSE);
	CheckDlgButton(hWnd, IDC_REMEMBER, set->rememberPassword ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_SHOW_SPLASH,set->showSplash?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd,IDC_CHECK_NEW,set->checkNewVersion?BST_CHECKED:BST_UNCHECKED);
	SetDlgItemText(hWnd,IDC_FONT_NAME,set->chatFontName);
}
//----------------------------------------------------------------------------
static void SetUserDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd, TEXT("set"));
	set->away_timeout = GetDlgItemInt(hWnd,IDC_SET_AWAYTIME,NULL,FALSE);
	set->rememberPassword = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_REMEMBER));
	set->passwordTime = GetDlgItemInt(hWnd,IDC_SET_REMEMBER,NULL,FALSE);
	set->autostart = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_STARTWIN));
	set->preventsleep = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_PREVENTSLEEP));
	set->openfirewall = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_FWPORTS));
	set->showSplash = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_SHOW_SPLASH));
	set->checkNewVersion = (BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_CHECK_NEW));
	GetDlgItemText(hWnd,IDC_FONT_NAME,set->chatFontName,sizeof(set->chatFontName)/sizeof(TCHAR));
}
//----------------------------------------------------------------------------
static HFONT CreateFontByName(LPTSTR name)
{
	TCHAR font[100] = {0};
	int pos;
	for (pos = cp_strlen(name); --pos;)
	{
		if (name[pos] == ',')
			break;
	}
	int size = _tcstol(&name[pos+1],NULL,10);
	if (!size) size = 10;
	cp_strncpy(font,name,pos);
	return CreateFont(-size,0,0,0,FW_DONTCARE,FALSE,FALSE,FALSE,DEFAULT_CHARSET,OUT_OUTLINE_PRECIS,CLIP_DEFAULT_PRECIS,CLEARTYPE_QUALITY,DEFAULT_PITCH,name);
}
//----------------------------------------------------------------------------
static void ChooseFont(HWND hWnd)
{
	CHOOSEFONT cf = {0};
	LOGFONT lf = {0};
	cf.lStructSize = sizeof(CHOOSEFONT);
	cf.hwndOwner = hWnd;
	cf.lpLogFont = &lf;
	if (ChooseFont(&cf))
	{
		settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
		_stprintf_s(set->chatFontName,TEXT("%s,%d"),lf.lfFaceName,cf.iPointSize/10);
		SetDlgItemText(hWnd,IDC_FONT_NAME,set->chatFontName);
	}
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetUserProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: SetUserInit(hWnd, (settings_t*)((PROPSHEETPAGE*)lParam)->lParam); return TRUE;
	case WM_DESTROY: SetUserDestroy(hWnd); return 0;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_REMEMBER: EnableWindow(GetDlgItem(hWnd,IDC_SET_REMEMBER),IsDlgButtonChecked(hWnd,IDC_REMEMBER)==BST_UNCHECKED); break;
		case IDC_FONT_CHANGE: ChooseFont(hWnd); break;
		}
	}
	return FALSE;
}
//----------------------------------------------------------------------------
typedef struct {
settings_t *set;
bool stop;
bool restart;
UINT WaveInNum;
HWND hwnd;
HANDLE thread;
} micThread_t;
//----------------------------------------------------------------------------
DWORD WINAPI micriphoneThread(micThread_t* mt)
{
	f64 avg;
	s16* buf;
	WAVEIN* wi;
	int updates;
	while(!mt->stop)
	{
		wi = new WAVEIN(mt->WaveInNum);
		if (!wi) 
		{
			Sleep(300); // something wrong, try again after some time
			continue;
		}
		mt->restart = false;
		updates = 0;
		avg = 0;
		while(!mt->stop && !mt->restart)
		{
			WaitForSingleObject(wi->rxEvent,500);
			while ((buf = (s16*)wi->GetBuffer()) != NULL)
			{
				for (int cnt = 0; cnt < WAVEIN_SAMPLES_PER_BUFFER; cnt++)
				{
					avg += ((f64)buf[cnt]*buf[cnt]*(1.0/32768.0/32768.0) - avg)*0.01;
				}
				if (!updates)
				{
					int level = (avg > 0) ? (int)(10.0*log(avg)/log(10.0)) : -60;
					if (level > 0) level = 0;
					if (level < -60) level = -60;

					PostMessage(GetDlgItem(mt->hwnd,IDC_MICPROGRESS),PBM_SETPOS,60+level,0);
					PostMessage(GetDlgItem(mt->hwnd,IDC_MICPROGRESS),PBM_SETBARCOLOR,0,level>=-11?RGB(0xFF,0,0):RGB(0,0xFF,0));
					PostMessage(mt->hwnd,WM_SHOW_LEVEL,0,level);
					updates = WAVEIN_FS/WAVEIN_SAMPLES_PER_BUFFER/4; // 4 updates per second
					setMicrophone(wi,mt->set->mic);
				}
				else
				{
					updates--;
				}
			}
		}
		delete wi;
	}
	return 0;
}
//----------------------------------------------------------------------------
DWORD soundThread(settings_t* set, LPCTSTR sound, UINT* volume)
{
    HRSRC  hResInfo;
	HGLOBAL hRes;
	WAVEOUT* wo;
	u8* wav_ptr;
	s16* buf;
	HeaderChunk_t head;
	FormatChunk_t fmt;
	DataChunk_t dat;
	u32 read_bytes = 0;
	u32 sent_bytes = 0;

	// parse wav file
    if (NULL == (hResInfo = FindResource(GetModuleHandle(NULL), sound, TEXT("WAVE")))) return 1;
    if (NULL == (hRes = LoadResource(GetModuleHandle(NULL), hResInfo))) return 1; 
	if (NULL == (wav_ptr = (u8*)LockResource(hRes))) return 1;
	memcpy(&head,wav_ptr,sizeof(HeaderChunk_t));
    if (head.chunkID != HeaderID || head.Format != HeaderFormatID) return 1;
	wav_ptr += sizeof(HeaderChunk_t);
	read_bytes += sizeof(HeaderChunk_t);
	memcpy(&fmt,wav_ptr,sizeof(FormatChunk_t));
    if (fmt.chunkID != FormatID || fmt.wFormatTag != 1 || fmt.wBitsPerSample != 16) return 1;
	wav_ptr += fmt.chunkSize + sizeof(DataChunk_t);
	read_bytes += fmt.chunkSize + sizeof(DataChunk_t);
   
    while (read_bytes < head.chunkSize-sizeof(HeaderChunk_t))
    {
		memcpy(&dat,wav_ptr,sizeof(DataChunk_t));
		if (dat.chunkID == DataID) break;
		wav_ptr += dat.chunkSize + sizeof(DataChunk_t);
		read_bytes += dat.chunkSize + sizeof(DataChunk_t);
	}
	if (dat.chunkID != DataID) return 1;
	if (NULL == (wo = new WAVEOUT(set->woNum,fmt.wChannels,fmt.dwSamplesPerSec))) return 1;

	wav_ptr += sizeof(DataChunk_t);
	u32 buf_size = wo->GetBufferSizeBytes();
	while(sent_bytes < dat.chunkSize)
	{
		WaitForSingleObject(wo->txEvent,500);
		while (NULL != (buf = (s16*)wo->GetBuffer()))
		{
			UINT cur_size = (sent_bytes + buf_size <= dat.chunkSize) ? buf_size : dat.chunkSize - sent_bytes;
			memcpy(buf,wav_ptr,cur_size);
			memset((u8*)buf + cur_size, 0, buf_size-cur_size);
			sent_bytes += cur_size;
			wav_ptr += cur_size;
			setVolume(wo,*volume);
		}
	}
    delete wo;
	return 0;
}

DWORD WINAPI speakerThread(settings_t* set)
{
	return soundThread(set,TEXT("waveTest"),&set->volume);
}

DWORD WINAPI ringThread(settings_t* set)
{
	return soundThread(set,TEXT("waveRing"),&set->ring_volume);
}
static DWORD selftestThread(HWND hWnd)
{
	settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
	static const u32 lenght_s = 20;
	WAVEIN* wi;
	WAVEOUT* wo;
	s16* buf = (s16*)LocalAlloc(LPTR,lenght_s*WAVEOUT_FS*sizeof(s16));
	s16* wbuf;
	UINT WaveInNum = ComboBox_GetCurSel(GetDlgItem(hWnd,IDC_MICROPHONE));
	if (WaveInNum) WaveInNum--;
	else WaveInNum = WAVE_MAPPER;
	u32 block_in, block_out;
	set->continue_rec = true;

	if (!buf)
	{
		MessageBox(hWnd,_T("Not enough memory for buffer"),_T("Warning"),MB_OK|MB_ICONWARNING);
		return -1;
	}
	wi = new WAVEIN(WaveInNum);
	if (!wi) 
	{
		LocalFree(buf);
		MessageBox(hWnd,_T("Can't open WAVEIN"),_T("Warning"),MB_OK|MB_ICONWARNING);
		return -2;
	}
	for (block_in = 0; block_in < lenght_s*WAVEOUT_FS/WAVEOUT_SAMPLES_PER_BUFFER && set->continue_rec;)
	{
		if (NULL != (wbuf = (s16*)wi->GetBuffer()))
		{
			memcpy(&buf[block_in*WAVEOUT_SAMPLES_PER_BUFFER],wbuf,WAVEOUT_SAMPLES_PER_BUFFER*sizeof(s16));
			block_in++;
			setMicrophone(wi,set->mic);
		}
		else
		{
			WaitForSingleObject(wi->rxEvent,500);
		}
	}
    delete wi;

	wo = new WAVEOUT(set->woNum);
	if (!wo) 
	{
		LocalFree(buf);
		MessageBox(hWnd,_T("Can't open WAVEOUT"),_T("Warning"),MB_OK|MB_ICONWARNING);
		return -3;
	}
	EndDialog(hWnd,0);
	for (block_out = 0; block_out < block_in; )
	{
		if (NULL != (wbuf = (s16*)wo->GetBuffer()))
		{
			memcpy(wbuf,&buf[block_out*WAVEOUT_SAMPLES_PER_BUFFER],WAVEOUT_SAMPLES_PER_BUFFER*sizeof(s16));
			block_out++;
			setVolume(wo,set->volume);
		}
		else
		{
			WaitForSingleObject(wo->txEvent,500);
		}
	}
    delete wo;
	LocalFree(buf);
	return 0;
}
static INT_PTR CALLBACK SpeackDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_INITDIALOG:
		SetProp(hWnd,_T("set"),(HANDLE)lParam);
		CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)selftestThread,hWnd,0,NULL));
		SetProp(hWnd,_T("tmo"),(HANDLE)20);
		SetTimer(hWnd,1,1000,NULL);
		return TRUE;
		//
	case WM_DESTROY:
		KillTimer(hWnd,1);
		RemoveProp(hWnd,_T("set"));
		return 0;
		//
	case WM_TIMER:
	{
		int tmo;
		TCHAR str[100];
		tmo = (int)GetProp(hWnd,_T("tmo"));
		if (tmo) tmo--;
		SetProp(hWnd,_T("tmo"),(HANDLE)tmo);
		_stprintf_s(str,_T("SPEAK %d sec"),tmo);
		SetDlgItemText(hWnd,IDC_STATUS,str);
		return 0;
	}
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			((settings_t*)GetProp(hWnd,_T("set")))->continue_rec = false;
			EndDialog(hWnd,0);
			return FALSE;
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------------

static void AudioSelfTest(HWND hWnd)
{
	DialogBoxParam(GetModuleHandle(NULL),_T("SpeakDialog"),hWnd,SpeackDlgProc,(LPARAM)GetProp(hWnd,_T("set")));
}

static void TestSpeaker(HWND hWnd, bool ring)
{
	settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
	set->woNum = ComboBox_GetCurSel(GetDlgItem(hWnd,IDC_SPEAKER));
	if (set->woNum) set->woNum--;
	else set->woNum = WAVE_MAPPER;

	if (ring) CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ringThread,set,0,NULL));
	else CloseHandle(CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)speakerThread,set,0,NULL));
}
//----------------------------------------------------------------------------
static void MicrophoneChanged(HWND hWnd)
{
	micThread_t* mt = (micThread_t*)GetProp(hWnd,TEXT("mic"));
	mt->WaveInNum = ComboBox_GetCurSel(GetDlgItem(hWnd,IDC_MICROPHONE));
	if (mt->WaveInNum) mt->WaveInNum--;
	else mt->WaveInNum = WAVE_MAPPER;
	mt->restart = true;
}
//----------------------------------------------------------------------------
static void SpeakerChanged(HWND hWnd)
{
}
//----------------------------------------------------------------------------
static void SetAudioInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd, TEXT("set"), set);
	// volume
	SendDlgItemMessage(hWnd,IDC_VOLUME,TBM_SETPOS,TRUE,set->volume);
	SendDlgItemMessage(hWnd,IDC_MICPROGRESS,PBM_SETRANGE32,0,60);
	SendDlgItemMessage(hWnd,IDC_MICLEVEL,TBM_SETPOS,TRUE,set->mic);
	SendDlgItemMessage(hWnd,IDC_RING_VOL,TBM_SETPOS,TRUE,set->ring_volume);
	CheckDlgButton(hWnd, IDC_MIC_AGC, set->mic_agc ? BST_CHECKED : BST_UNCHECKED);
	CheckRadioButton(hWnd,IDC_AUD_LOW,IDC_AUD_HIGH,IDC_AUD_LOW+set->audio_quality);
	// enum wavein
	ComboBox_AddString(GetDlgItem(hWnd,IDC_MICROPHONE),TEXT("default"));
	for (UINT i = 0; i < waveInGetNumDevs(); i++)
	{
		WAVEINCAPS cap;
		if (MMSYSERR_NOERROR == waveInGetDevCaps(i,&cap,sizeof(cap)))
		{
			ComboBox_AddString(GetDlgItem(hWnd,IDC_MICROPHONE),cap.szPname);
		}
	}
	if (CB_ERR == ComboBox_SelectString(GetDlgItem(hWnd,IDC_MICROPHONE),0,set->microphone))
	{
		ComboBox_SetCurSel(GetDlgItem(hWnd,IDC_MICROPHONE),0);
	}
	// enum waveout
	ComboBox_AddString(GetDlgItem(hWnd,IDC_SPEAKER),TEXT("default"));
	for (UINT i = 0; i < waveOutGetNumDevs(); i++)
	{
		WAVEOUTCAPS cap;
		if (MMSYSERR_NOERROR == waveOutGetDevCaps(i,&cap,sizeof(cap)))
		{
			ComboBox_AddString(GetDlgItem(hWnd,IDC_SPEAKER),cap.szPname);
		}
	}
	if (CB_ERR == ComboBox_SelectString(GetDlgItem(hWnd,IDC_SPEAKER),0,set->speaker))
	{
		ComboBox_SetCurSel(GetDlgItem(hWnd,IDC_SPEAKER),0);
	}
	//
	micThread_t* mt = (micThread_t*)LocalAlloc(LPTR,sizeof(micThread_t));
	mt->set = set;
	mt->WaveInNum = ComboBox_GetCurSel(GetDlgItem(hWnd,IDC_MICROPHONE));
	if (mt->WaveInNum) mt->WaveInNum--;
	else mt->WaveInNum = WAVE_MAPPER;
	mt->hwnd = hWnd;
	mt->thread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)micriphoneThread,mt,0,NULL);
	SetProp(hWnd,TEXT("mic"),mt);
}
//----------------------------------------------------------------------------
static int getQuality(HWND hWnd, UINT id)
{
	if (BST_CHECKED == IsDlgButtonChecked(hWnd, id)) return 0;
	else if (BST_CHECKED == IsDlgButtonChecked(hWnd, id+1)) return 1;
	else if (BST_CHECKED == IsDlgButtonChecked(hWnd, id+2)) return 2;
	return 3;
}
//----------------------------------------------------------------------------
static void SetAudioDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd, TEXT("set"));
	micThread_t* mt = (micThread_t*)RemoveProp(hWnd,TEXT("mic"));
	mt->stop = true;
	WaitForSingleObject(mt->thread,5000);
	CloseHandle(mt->thread);
	LocalFree(mt);
	GetDlgItemText(hWnd,IDC_MICROPHONE,set->microphone,MAXPNAMELEN);
	GetDlgItemText(hWnd,IDC_SPEAKER,set->speaker,MAXPNAMELEN);
	set->mic_agc = (BST_CHECKED == IsDlgButtonChecked(hWnd, IDC_MIC_AGC));
	set->audio_quality = getQuality(hWnd,IDC_AUD_LOW);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetAudioProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetAudioInit(hWnd, (settings_t*)((PROPSHEETPAGE*)lParam)->lParam);
		return TRUE;
		//
	case WM_DESTROY:
		SetAudioDestroy(hWnd);
		return 0;
		//
	case WM_HSCROLL:
		if ((HWND)lParam == GetDlgItem(hWnd,IDC_VOLUME))
		{
			settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
			set->volume = (UINT)SendDlgItemMessage(hWnd,IDC_VOLUME,TBM_GETPOS,0,0);
		}
		if ((HWND)lParam == GetDlgItem(hWnd,IDC_MICLEVEL))
		{
			settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
			set->mic = (UINT)SendDlgItemMessage(hWnd,IDC_MICLEVEL,TBM_GETPOS,0,0);
		}
		if ((HWND)lParam == GetDlgItem(hWnd,IDC_RING_VOL))
		{
			settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
			set->ring_volume = (UINT)SendDlgItemMessage(hWnd,IDC_RING_VOL,TBM_GETPOS,0,0);
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_TEST_SPEAKER:
		case IDC_TEST_RING:
			TestSpeaker(hWnd,LOWORD(wParam)==IDC_TEST_RING);
			return TRUE;
		case IDC_MICROPHONE:
			if (HIWORD(wParam)==CBN_SELCHANGE) MicrophoneChanged(hWnd);
			return TRUE;
		case IDC_SPEAKER:
			if (HIWORD(wParam)==CBN_SELCHANGE) SpeakerChanged(hWnd);
			return TRUE;
		case IDC_MIC_AGC:
			apply_AGC(BST_CHECKED==IsDlgButtonChecked(hWnd,IDC_MIC_AGC));
			return TRUE;
		case IDC_AUD_LOW:
		case IDC_AUD_NORMAL:
		case IDC_AUD_HIGH:
			apply_AQ(getQuality(hWnd,IDC_AUD_LOW));
			break;
		case IDC_AUDIO_TEST:
			AudioSelfTest(hWnd);
			break;
		}
		return FALSE;
	case WM_SHOW_LEVEL:
	{
		TCHAR str[10];
		_stprintf_s(str,TEXT("%d dB"),lParam);
		SetDlgItemText(hWnd,IDC_MIC_DB,str);
		return FALSE;
	}
	}
	return FALSE;
}
//-----------------------------------------------------------------------
static void SetVideoInit(HWND hWnd, settings_t* set)
{
	SetProp(hWnd, TEXT("set"), set);
	CheckRadioButton(hWnd,IDC_VID_LOW,IDC_VID_HIGH,IDC_VID_LOW+set->video_quality);
	CheckRadioButton(hWnd,IDC_VID_RES1,IDC_VID_RES4,IDC_VID_RES1+set->video_resolution);
	fill_video_inputs(GetDlgItem(hWnd,IDC_VIDEO_IN));
	if (CB_ERR == ComboBox_SelectString(GetDlgItem(hWnd,IDC_VIDEO_IN),-1,set->video))
		ComboBox_SetCurSel(GetDlgItem(hWnd,IDC_VIDEO_IN),0);
	GetDlgItemText(hWnd,IDC_VIDEO_IN,set->video,array_size(set->video));
	ShowWindow(GetDlgItem(hWnd,IDC_VID_RES1),SW_HIDE);
	ShowWindow(GetDlgItem(hWnd,IDC_VID_RES2),SW_HIDE);
	ShowWindow(GetDlgItem(hWnd,IDC_VID_RES3),SW_HIDE);
	ShowWindow(GetDlgItem(hWnd,IDC_VID_RES4),SW_HIDE);
	void* vid = video_create();
	SetProp(hWnd,_T("video"),vid);
	video_start(vid,NULL,set->video);
}
//----------------------------------------------------------------------------
static void SetVideoDestroy(HWND hWnd)
{
	settings_t* set = (settings_t*)RemoveProp(hWnd, TEXT("set"));
	set->video_quality = getQuality(hWnd,IDC_VID_LOW);
	set->video_resolution = getQuality(hWnd,IDC_VID_RES1);
	GetDlgItemText(hWnd,IDC_VIDEO_IN,set->video,array_size(set->video));
	void* vid = RemoveProp(hWnd,_T("video"));
	video_stop(vid);
	video_delete(vid);
}
//----------------------------------------------------------------------------
static INT_PTR CALLBACK SetVideoProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetVideoInit(hWnd, (settings_t*)((PROPSHEETPAGE*)lParam)->lParam);
		return TRUE;
		//
	case WM_DESTROY:
		SetVideoDestroy(hWnd);
		return 0;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_VIDEO_IN:
			if (HIWORD(wParam)==CBN_SELCHANGE) {
				settings_t* set = (settings_t*)GetProp(hWnd, TEXT("set"));
				GetDlgItemText(hWnd,IDC_VIDEO_IN,set->video,array_size(set->video));
				void* vid = GetProp(hWnd,_T("video"));
				video_stop(vid);
				video_start(vid,NULL,set->video);
			}
			return TRUE;
		}
	}
	return FALSE;
}
//-----------------------------------------------------------------------
static void sturtupZAS(int startup)
{
	HKEY key;
	if (ERROR_SUCCESS == RegOpenKeyEx(HKEY_CURRENT_USER,TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Run"),0,KEY_ALL_ACCESS,&key))
	{
		if (startup)
		{
			TCHAR fileName[MAX_PATH];
			GetModuleFileName(NULL,&fileName[1],MAX_PATH-1);
			fileName[0] = '\"';
			cp_strcchcat(fileName,MAX_PATH,TEXT("\""));
			RegSetValueEx(key,TEXT("ZAS"),0,REG_SZ,(BYTE*)fileName,(DWORD)((cp_strlen(fileName)+1)*sizeof(TCHAR)));
		}
		else
		{
			RegDeleteValue(key,TEXT("ZAS"));
		}
		RegCloseKey(key);
	}
}
//----------------------------------------------------------------------------
#define PAGES_NUM	6
//----------------------------------------------------------------------------
void SettingDialog(HWND hWnd, settings_t* set, USERS* usr, KADEMLIA* kad)
{
	PROPSHEETPAGE psp[PAGES_NUM];
	PROPSHEETHEADER psh;
	settings_t loc_setttings = *set;
	
	psp[0].dwSize = sizeof(PROPSHEETPAGE);
	psp[0].dwFlags = PSP_USETITLE;
	psp[0].hInstance = GetModuleHandle(NULL);
	psp[0].pszTemplate = TEXT("SettingsFiles");
	psp[0].pfnDlgProc = SetFileProc;
	psp[0].pszTitle = TEXT("File");
	psp[0].lParam = (LPARAM)&loc_setttings;
	psp[0].pfnCallback = NULL;
	psp[0].pszIcon = NULL;
	//
	psp[1].dwSize = sizeof(PROPSHEETPAGE);
	psp[1].dwFlags = PSP_USETITLE;
	psp[1].hInstance = GetModuleHandle(NULL);
	psp[1].pszTemplate = TEXT("SettingsNetwork");
	psp[1].pfnDlgProc = SetNetworkProc;
	psp[1].pszTitle = TEXT("Network");
	psp[1].lParam = (LPARAM)&loc_setttings;
	psp[1].pfnCallback = NULL;
	psp[1].pszIcon = NULL;
	//
	psp[2].dwSize = sizeof(PROPSHEETPAGE);
	psp[2].dwFlags = PSP_USETITLE;
	psp[2].hInstance = GetModuleHandle(NULL);
	psp[2].pszTemplate = TEXT("SettingsPrivacy");
	psp[2].pfnDlgProc = SetPrivacyProc;
	psp[2].pszTitle = TEXT("Privacy");
	psp[2].lParam = (LPARAM)&loc_setttings;
	psp[2].pfnCallback = NULL;
	psp[2].pszIcon = NULL;
	//
	psp[3].dwSize = sizeof(PROPSHEETPAGE);
	psp[3].dwFlags = PSP_USETITLE;
	psp[3].hInstance = GetModuleHandle(NULL);
	psp[3].pszTemplate = TEXT("SettingsUser");
	psp[3].pfnDlgProc = SetUserProc;
	psp[3].pszTitle = TEXT("User");
	psp[3].lParam = (LPARAM)&loc_setttings;
	psp[3].pfnCallback = NULL;
	psp[3].pszIcon = NULL;
	//
	psp[4].dwSize = sizeof(PROPSHEETPAGE);
	psp[4].dwFlags = PSP_USETITLE;
	psp[4].hInstance = GetModuleHandle(NULL);
	psp[4].pszTemplate = TEXT("SettingsAudio");
	psp[4].pfnDlgProc = SetAudioProc;
	psp[4].pszTitle = TEXT("Audio");
	psp[4].lParam = (LPARAM)&loc_setttings;
	psp[4].pfnCallback = NULL;
	psp[4].pszIcon = NULL;
	//
	psp[5].dwSize = sizeof(PROPSHEETPAGE);
	psp[5].dwFlags = PSP_USETITLE;
	psp[5].hInstance = GetModuleHandle(NULL);
	psp[5].pszTemplate = TEXT("SettingsVideo");
	psp[5].pfnDlgProc = SetVideoProc;
	psp[5].pszTitle = TEXT("Video");
	psp[5].lParam = (LPARAM)&loc_setttings;
	psp[5].pfnCallback = NULL;
	psp[5].pszIcon = NULL;
	//
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_PROPSHEETPAGE | PSH_NOCONTEXTHELP | PSH_NOAPPLYNOW;
	psh.hwndParent = hWnd;
	psh.hInstance = GetModuleHandle(NULL);
	psh.pszCaption = TEXT("Settings");
	psh.nPages = PAGES_NUM;
	psh.nStartPage = 0;
	psh.ppsp = (LPCPROPSHEETPAGE) &psp;
	psh.pfnCallback = NULL;
	if(PropertySheet(&psh)) // copy settings if update is successful
	{
		if ((set->enableUPnP!=loc_setttings.enableUPnP)||(set->randomPort!=loc_setttings.randomPort)||(set->remotePort!=loc_setttings.remotePort)||(set->localPort!=loc_setttings.localPort)||(set->nextAvailablePort!=loc_setttings.nextAvailablePort))
		{
			if (loc_setttings.enableUPnP)
			{
				if (!setupUPnP(&loc_setttings,0))
				{
					MessageBox(hWnd,TEXT("Can't setup UPnP"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
				}
			}
			else
			{
				if (!setupUPnP(set,1))
				{
					MessageBox(hWnd,TEXT("Can't disable UPnP"),TEXT("Warning"),MB_OK|MB_ICONWARNING);
				}
			}
		}
		if (_tcscmp(set->keyFileName,loc_setttings.keyFileName))
		{
			LoadKey(hWnd, loc_setttings.keyFileName, getSecretData());
		}
		if (_tcscmp(set->logFileName,loc_setttings.logFileName) || (set->enableLogFile^loc_setttings.enableLogFile))
		{
			SendMessage(g_logWnd,WM_APPLY_LOG_SETTINGS,0,(LPARAM)&loc_setttings);
		}
		if (_tcscmp(set->userFileName,loc_setttings.userFileName))
		{
			usr->load_nodes(kad,loc_setttings.userFileName);
		}
		apply_AGC(loc_setttings.mic_agc);
		apply_AQ(loc_setttings.audio_quality);
		*set = loc_setttings;
		// apply new settings
		SetWaveParam(set);
		sturtupZAS(set->autostart);
		SetThreadExecutionState(ES_CONTINUOUS|(set->preventsleep?ES_SYSTEM_REQUIRED:0));
		if (set->openfirewall) EnablePortInFirewall(set->localPort);
		else DisablePortInFirewall(set->localPort);
		// Delete old font and create new font
		DeleteObject(getChatFont());
		setChatFont(CreateFontByName(set->chatFontName));
		apply_Font();
	}
}
//-----------------------------------------------------------------------
static void getSettingsFileName(LPTSTR fileName)
{
	GetModuleFileName(NULL,fileName,MAX_PATH);
	CopyMemory(&fileName[_tcslen(fileName)-3],TEXT("set"),3*sizeof(TCHAR));
}
//-----------------------------------------------------------------------
void getModuleDirectory(LPTSTR dir)
{
	size_t i;
	GetModuleFileName(NULL,dir,MAX_PATH);
	for (i = _tcslen(dir); --i;)
		if (dir[i] == '\\') break;
	dir[i] = 0;
}
//-----------------------------------------------------------------------
void LoadSettings(HWND hWnd, settings_t* set)
{
	TCHAR moduleName[MAX_PATH];
	TCHAR fileName[MAX_PATH];
	TCHAR defName[MAX_PATH];
	getSettingsFileName(fileName);
	GetModuleFileName(NULL,moduleName,MAX_PATH);
	for (size_t i = cp_strlen(moduleName);i--;)
	{
		if (moduleName[i] == '\\')
		{
			moduleName[i] = 0;
			break;
		}
	}

	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\zas.key"));
	if (!PathFileExists(fileName))
	{
		TCHAR msg[MAX_PATH+100];
		_stprintf_s(msg,MAX_PATH+100,TEXT("Can't find settings file:\n%s\nDefault settings will be used.\nDo you want to generate key file now?"),fileName);
		if (IDYES == MessageBox(hWnd,msg,TEXT("Warning"),MB_YESNO|MB_ICONWARNING))
		{
			GenerateKey(hWnd,defName);
		}
	}
	GetPrivateProfileString(TEXT("Files"),TEXT("Key"),defName,set->keyFileName,MAX_PATH,fileName);
	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\user.xml"));
	GetPrivateProfileString(TEXT("Files"),TEXT("Users"),defName,set->userFileName,MAX_PATH,fileName);
	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\likes.xml"));
	GetPrivateProfileString(TEXT("Files"),TEXT("Likes"),defName,set->likesFileName,MAX_PATH,fileName);
	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\log.txt"));
	GetPrivateProfileString(TEXT("Files"),TEXT("Log"),defName,set->logFileName,MAX_PATH,fileName);
	set->enableLogFile= GetPrivateProfileInt(TEXT("Files"),TEXT("EnableLog"),0,fileName);
	set->log_size= GetPrivateProfileInt(TEXT("Files"),TEXT("LogSize"),100,fileName);
	GetPrivateProfileString(TEXT("Files"), TEXT("FileTransferPath"),moduleName, set->ftPath, MAX_PATH, fileName);
	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\downloads.xml"));
	GetPrivateProfileString(TEXT("Files"),TEXT("Downloads"),defName,set->downloadsFileName,MAX_PATH,fileName);
	cp_snprintf(defName,MAX_PATH,TEXT("%s%s"),moduleName,TEXT("\\seed.bin"));
	GetPrivateProfileString(TEXT("Files"),TEXT("Seed"),defName,set->seedFileName,MAX_PATH,fileName);
	set->reciveFiles=GetPrivateProfileInt(TEXT("Files"),TEXT("AutoReceiveFiles"),0,fileName);


	set->passwordTime = GetPrivateProfileInt(TEXT("Password"),TEXT("Time"),0,fileName);
	set->rememberPassword = GetPrivateProfileInt(TEXT("Password"),TEXT("Remember"),0,fileName);
	if (set->rememberPassword)
	{
		TCHAR val[2*password_max_length+10];
		u8 pass[password_max_length];
		GetPrivateProfileString(TEXT("Password"),TEXT("Pass"),TEXT(""),val,2*password_max_length+10,fileName);
		size_t len = cp_strlen(val);
		if (len)
		{
			for(size_t i = len/2; i--;)
			{
				pass[len/2-i-1] = (_tcstol(&val[2*i],NULL,16)^(i+(i<<4)))&0xFF;
				val[2*i] = 0;
			}
			set->passwordTime = 0;
			setPassword(pass,(u32)(len/2),set);
		}
	}

	set->localPort = GetPrivateProfileInt(TEXT("Network"),TEXT("LocalPort"),31265,fileName);
	set->remotePort = GetPrivateProfileInt(TEXT("Network"),TEXT("RemotePort"),31265,fileName);
	set->enableUPnP = GetPrivateProfileInt(TEXT("Network"),TEXT("EnableUPnP"),1,fileName);
	set->removeUPnPonExit = GetPrivateProfileInt(TEXT("Network"),TEXT("CancelUPnPonExit"),1,fileName);
	set->nextAvailablePort = GetPrivateProfileInt(TEXT("Network"),TEXT("NextAvailablePort"),1,fileName);
	set->randomPort = GetPrivateProfileInt(TEXT("Network"),TEXT("RandomPort"),0,fileName);
	set->encapsulation = GetPrivateProfileInt(TEXT("Network"),TEXT("Encapsulation"),0,fileName);
	set->obfuscationEnabled = GetPrivateProfileInt(TEXT("Network"),TEXT("Obfuscation"),1,fileName);
	set->replaceLocalIP = GetPrivateProfileInt(TEXT("Network"),TEXT("ReplaceLocalIP"),1,fileName);

	set->allowAnyone = GetPrivateProfileInt(TEXT("Privacy"), TEXT("AllowAnyone"), 1, fileName);
	set->allowConnection = GetPrivateProfileInt(TEXT("Privacy"), TEXT("AllowConnectionRequest"), 1, fileName);
	set->acceptFT = GetPrivateProfileInt(TEXT("Privacy"), TEXT("AcceptFileTransfer"), 1, fileName);

	set->away_timeout = GetPrivateProfileInt(TEXT("System"), TEXT("AwayTimeout"), 15, fileName);
	set->autostart = GetPrivateProfileInt(TEXT("System"),TEXT("Autostart"),0,fileName);
	set->preventsleep = GetPrivateProfileInt(TEXT("System"),TEXT("PreventSleep"),0,fileName);
	set->openfirewall = GetPrivateProfileInt(TEXT("System"),TEXT("OpenFirewall"),0,fileName);
	set->showSplash = GetPrivateProfileInt(TEXT("System"),TEXT("ShowSplashWindow"),1,fileName);
	set->checkNewVersion = GetPrivateProfileInt(TEXT("System"),TEXT("CheckNewVersion"),1,fileName);
	if (set->checkNewVersion)
	{
		TCHAR time_str[20];
		cp_time_t last_time,cur_time;
		cp_get_time(&cur_time);
		GetPrivateProfileString(TEXT("System"), TEXT("CheckNewVersionLast"), TEXT(""), time_str, 20, fileName);
		if (!str2time(time_str,&last_time) || (cur_time-last_time > 7*24*3600*10000000ull))
		{
			time2str(cur_time,time_str);
			WritePrivateProfileString(TEXT("System"), TEXT("CheckNewVersionLast"), time_str, fileName);
			CheckNewVersion(hWnd);
		}
	}
	GetPrivateProfileString(TEXT("System"), TEXT("Chat Font"), TEXT("Times New Roman,12"), set->chatFontName, sizeof(set->chatFontName)/sizeof(TCHAR), fileName);

	GetPrivateProfileString(TEXT("Audio"), TEXT("Microphone"), TEXT("default"), set->microphone, MAXPNAMELEN, fileName);
	GetPrivateProfileString(TEXT("Audio"), TEXT("Speaker"), TEXT("default"), set->speaker, MAXPNAMELEN, fileName);
	set->volume = GetPrivateProfileInt(TEXT("Audio"), TEXT("Volume"), 66, fileName);
	set->mic = GetPrivateProfileInt(TEXT("Audio"), TEXT("MicLevel"), 66, fileName);
	set->mic_agc = GetPrivateProfileInt(TEXT("Audio"), TEXT("AGC"), 1, fileName);
	set->audio_quality = GetPrivateProfileInt(TEXT("Audio"), TEXT("Quality"), 1, fileName);
	set->ring_volume = GetPrivateProfileInt(TEXT("Audio"), TEXT("RingVolume"), 66, fileName);

	set->video_quality = GetPrivateProfileInt(TEXT("Video"), TEXT("VideoQuality"), 1, fileName);
	set->video_resolution = GetPrivateProfileInt(TEXT("Video"), TEXT("VideoResolution"), 1, fileName);
	GetPrivateProfileString(TEXT("Video"), TEXT("Video"), TEXT(""), set->video, 64, fileName);

	set->showSessionDebug = GetPrivateProfileInt(TEXT("Debug"), TEXT("Session"), 1, fileName);

	sturtupZAS(set->autostart); // update path if it's changed
	SetThreadExecutionState(ES_CONTINUOUS|(set->preventsleep?ES_SYSTEM_REQUIRED:0));
	if (set->openfirewall)EnablePortInFirewall(set->localPort);
	setChatFont(CreateFontByName(set->chatFontName));
}
//-----------------------------------------------------------------------
void SaveSettings(HWND hWnd, settings_t* set)
{
	TCHAR val[100];
	TCHAR fileName[MAX_PATH];
	getSettingsFileName(fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Key"),set->keyFileName,fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Users"),set->userFileName,fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Likes"),set->likesFileName,fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Log"),set->logFileName,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->enableLogFile);
	WritePrivateProfileString(TEXT("Files"),TEXT("EnableLog"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->log_size);
	WritePrivateProfileString(TEXT("Files"),TEXT("LogSize"),val,fileName);
	WritePrivateProfileString(TEXT("Files"), TEXT("FileTransferPath"), set->ftPath, fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Downloads"),set->downloadsFileName,fileName);
	WritePrivateProfileString(TEXT("Files"),TEXT("Seed"),set->seedFileName,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->reciveFiles);
	WritePrivateProfileString(TEXT("Files"),TEXT("AutoReceiveFiles"),val,fileName);

	_stprintf_s(val,100,TEXT("%d"),set->passwordTime);
	WritePrivateProfileString(TEXT("Password"),TEXT("Time"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->rememberPassword);
	WritePrivateProfileString(TEXT("Password"),TEXT("Remember"),val,fileName);
	if (set->rememberPassword)
	{
		TCHAR p[2*password_max_length+10];
		u8 pass[password_max_length];
		u32 len = getPassword(pass);
		if (len)
		{
			for(u32 i = 0; i < len; i++)
			{
				_stprintf_s(&p[2*i],2*password_max_length+10-2*i,TEXT("%02X"),(pass[len-i-1]^(i+(i<<4)))&0xFF);
			}
		}
		WritePrivateProfileString(TEXT("Password"),TEXT("Pass"),p,fileName);
	}
	else
	{
		WritePrivateProfileString(TEXT("Password"),TEXT("Pass"),TEXT(""),fileName);
	}

	_stprintf_s(val,100,TEXT("%d"),set->localPort);
	WritePrivateProfileString(TEXT("Network"),TEXT("LocalPort"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->remotePort);
	WritePrivateProfileString(TEXT("Network"),TEXT("RemotePort"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->enableUPnP);
	WritePrivateProfileString(TEXT("Network"),TEXT("EnableUPnP"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->removeUPnPonExit);
	WritePrivateProfileString(TEXT("Network"),TEXT("CancelUPnPonExit"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->nextAvailablePort);
	WritePrivateProfileString(TEXT("Network"),TEXT("NextAvailablePort"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->randomPort);
	WritePrivateProfileString(TEXT("Network"),TEXT("RandomPort"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->encapsulation);
	WritePrivateProfileString(TEXT("Network"),TEXT("Encapsulation"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->obfuscationEnabled);
	WritePrivateProfileString(TEXT("Network"),TEXT("Obfuscation"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->replaceLocalIP);
	WritePrivateProfileString(TEXT("Network"),TEXT("ReplaceLocalIP"),val,fileName);

	_stprintf_s(val, 100, TEXT("%d"), set->allowAnyone);
	WritePrivateProfileString(TEXT("Privacy"), TEXT("AllowAnyone"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->allowConnection);
	WritePrivateProfileString(TEXT("Privacy"), TEXT("AllowConnectionRequest"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->acceptFT);
	WritePrivateProfileString(TEXT("Privacy"), TEXT("AcceptFileTransfer"), val, fileName);

	_stprintf_s(val, 100, TEXT("%d"), set->away_timeout);
	WritePrivateProfileString(TEXT("System"), TEXT("AwayTimeout"), val, fileName);
	_stprintf_s(val,100,TEXT("%d"),set->autostart);
	WritePrivateProfileString(TEXT("System"),TEXT("Autostart"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->preventsleep);
	WritePrivateProfileString(TEXT("System"),TEXT("PreventSleep"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->openfirewall);
	WritePrivateProfileString(TEXT("System"),TEXT("OpenFirewall"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->showSplash);
	WritePrivateProfileString(TEXT("System"),TEXT("ShowSplashWindow"),val,fileName);
	_stprintf_s(val,100,TEXT("%d"),set->checkNewVersion);
	WritePrivateProfileString(TEXT("System"),TEXT("CheckNewVersion"),val,fileName);
	WritePrivateProfileString(TEXT("System"), TEXT("Chat Font"), set->chatFontName, fileName);
	

	WritePrivateProfileString(TEXT("Audio"), TEXT("Microphone"), set->microphone, fileName);
	WritePrivateProfileString(TEXT("Audio"), TEXT("Speaker"), set->speaker, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->volume);
	WritePrivateProfileString(TEXT("Audio"), TEXT("Volume"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->mic);
	WritePrivateProfileString(TEXT("Audio"), TEXT("MicLevel"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->mic_agc);
	WritePrivateProfileString(TEXT("Audio"), TEXT("AGC"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->audio_quality);
	WritePrivateProfileString(TEXT("Audio"), TEXT("Quality"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->ring_volume);
	WritePrivateProfileString(TEXT("Audio"), TEXT("RingVolume"), val, fileName);

	_stprintf_s(val, 100, TEXT("%d"), set->video_quality);
	WritePrivateProfileString(TEXT("Video"), TEXT("VideoQuality"), val, fileName);
	_stprintf_s(val, 100, TEXT("%d"), set->video_resolution);
	WritePrivateProfileString(TEXT("Video"), TEXT("VideoResolution"), val, fileName);
	WritePrivateProfileString(TEXT("Video"), TEXT("Video"), set->video, fileName);

	_stprintf_s(val, 100, TEXT("%d"), set->showSessionDebug);
	WritePrivateProfileString(TEXT("Debug"), TEXT("Session"), val, fileName);
}
//-----------------------------------------------------------------------
bool GetSaveChat(const u32* id)
{
	TCHAR str[100];
	TCHAR fileName[MAX_PATH];
	getSettingsFileName(fileName);
	id2str(id,str);
	return (1==GetPrivateProfileInt(TEXT("SaveChat"), str, 1, fileName));
}
//-----------------------------------------------------------------------
void SetSaveChat(const u32* id,bool save)
{
	TCHAR val[10];
	TCHAR str[100];
	TCHAR fileName[MAX_PATH];
	getSettingsFileName(fileName);
	id2str(id,str);
	_stprintf_s(val, TEXT("%d"), save);
	WritePrivateProfileString(TEXT("SaveChat"), str, val, fileName);
}
//-----------------------------------------------------------------------
VOID URL2IP(LPCTSTR url, u32* ip)
{
	PDNS_RECORD pDnsRecord;
	*ip = 0;
	if (DNS_RCODE_NOERROR == DnsQuery(url,DNS_TYPE_A,DNS_QUERY_NO_NETBT|DNS_QUERY_NO_HOSTS_FILE,NULL,&pDnsRecord,NULL))
	{
		for (PDNS_RECORD rr = pDnsRecord; rr; rr = rr->pNext)
		{
			if (rr->wType == DNS_TYPE_A)
			{
				*ip = ntohl(rr->Data.A.IpAddress);
				break;
			}
		}
		DnsRecordListFree(pDnsRecord,DnsFreeRecordList);
	}
}
//----------------------------------------------------------------------------
typedef struct {
	TCHAR version[10];
	TCHAR date[20];
	TCHAR url[L_MAX_URL_LENGTH];
	int build;
} ver_t;
//----------------------------------------------------------------------------
static INT_PTR CALLBACK CheckNewVersionProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
	{
		TCHAR str[500];
		LITEM link;
		ver_t* ver = (ver_t*)lParam;
		_stprintf_s(str,TEXT("New version is available\nZAS version %s build %d (released %s)"),ver->version,ver->build,ver->date);
		SetDlgItemText(hWnd,IDC_VERSION_TEXT,str);
		link.mask = LIF_ITEMINDEX|LIF_URL;
		link.iLink = 0;
		cp_strcpy_s(link.szUrl,ver->url);
		_stprintf_s(str,TEXT("For more information, <A HREF=\"%s\">click here</A>"),ver->url);
		SetDlgItemText(hWnd,IDC_URL,str);
		SendDlgItemMessage(hWnd,IDC_URL,LM_SETITEM,0,(LPARAM)&link);
		return TRUE;
	}
	case WM_DESTROY:
		return 0;
	case WM_CLOSE:
		EndDialog(hWnd, 0);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hWnd, 0);
			return TRUE;
		}
		return FALSE;
	case WM_NOTIFY:
		if(((NMHDR*)lParam)->code == NM_CLICK)
		{
            PNMLINK pNMLink = (PNMLINK)lParam;
            ShellExecute(NULL, TEXT("open"), pNMLink->item.szUrl, NULL, NULL, SW_SHOW);
		}
		return FALSE;
	}
	return FALSE;
}
//----------------------------------------------------------------------
void CheckNewVersion(HWND hWnd, bool show_same)
{
	TCHAR filename[MAX_PATH];
	if (S_OK != URLDownloadToCacheFile(NULL,TEXT("http://www.zas-comm.ru/version.txt"),filename,MAX_PATH,0,NULL)) return;
	int build = GetPrivateProfileInt(TEXT("ZAS_Version"), TEXT("Build"), BUILD_NUM, filename);
	if (BUILD_NUM < build)
	{
		ver_t ver;
		ver.build = build;
		GetPrivateProfileString(TEXT("ZAS_Version"), TEXT("Version"), TEXT("unknown"), ver.version, sizeof(ver.version)/sizeof(TCHAR), filename);
		GetPrivateProfileString(TEXT("ZAS_Version"), TEXT("DownloadPath"), TEXT("http://www.zas-comm.ru/downloads.html"), ver.url, sizeof(ver.url)/sizeof(TCHAR), filename);
		GetPrivateProfileString(TEXT("ZAS_Version"), TEXT("Date"), TEXT("today"), ver.date, sizeof(ver.date)/sizeof(TCHAR), filename);
		DialogBoxParam(GetModuleHandle(NULL), TEXT("CheckNewVersionDialog"), hWnd, CheckNewVersionProc, (LPARAM)&ver);
	}
	else if (show_same)
	{
		MessageBox(hWnd,TEXT("You have the current version"),TEXT("Version"),MB_ICONINFORMATION|MB_OK);
	}
}
//-----------------------------------------------------------------------
VOID SaveWindowPosition(LPTSTR name, WINDOWPLACEMENT* pos)
{
	TCHAR val[100];
	TCHAR fileName[MAX_PATH];
	getSettingsFileName(fileName);

	_stprintf_s(val, TEXT("%d,%d,%d,%d,%d"), pos->rcNormalPosition.left,pos->rcNormalPosition.top,pos->rcNormalPosition.right,pos->rcNormalPosition.bottom,pos->showCmd);
	WritePrivateProfileString(TEXT("Position"),name,val,fileName);
}
//-----------------------------------------------------------------------
VOID LoadWindowPosition(LPTSTR name, WINDOWPLACEMENT* pos)
{
	TCHAR val[100];
	TCHAR fileName[MAX_PATH];
	int param[5];
	getSettingsFileName(fileName);

	GetPrivateProfileString(TEXT("Position"), name, TEXT("0,0,0,0,0"), val, 100, fileName);
	ZeroMemory(pos,sizeof(WINDOWPLACEMENT));
	if (5 == _stscanf_s(val,TEXT("%d,%d,%d,%d,%d"),&param[0],&param[1],&param[2],&param[3],&param[4]))
	{
		RECT vs;
		pos->rcNormalPosition.left = param[0];
		pos->rcNormalPosition.top = param[1];
		pos->rcNormalPosition.right = param[2];
		pos->rcNormalPosition.bottom = param[3];
		pos->showCmd = param[4];
		vs.left = GetSystemMetrics(SM_XVIRTUALSCREEN);
		vs.top = GetSystemMetrics(SM_YVIRTUALSCREEN);
		vs.right = vs.left + GetSystemMetrics(SM_CXVIRTUALSCREEN);
		vs.bottom = vs.top + GetSystemMetrics(SM_CYVIRTUALSCREEN);
		if (pos->rcNormalPosition.left < vs.left || pos->rcNormalPosition.right > vs.right ||
			pos->rcNormalPosition.top < vs.top || pos->rcNormalPosition.bottom > vs.bottom)
		{
			ZeroMemory(pos,sizeof(WINDOWPLACEMENT));
		}
	}
	pos->length = sizeof(WINDOWPLACEMENT);
}
//-----------------------------------------------------------------------
