#include "..\UPnP\miniwget.h"
#include "..\UPnP\miniupnpc.h"
#include "..\UPnP\upnpcommands.h"
#include "..\UPnP\upnperrors.h"
#include <windows.h>
#include <stdio.h>
#include <Iphlpapi.h>
#include "types.h"
#include "log.h"
#include "upnp.h"
#include "random.h"
//-----------------------------------------
static LPTSTR module = TEXT("upnp");
static u32 ip = (127<<24)+1;
static u32 int_srv_ip = 0;
static u32 int_ip = 0;
//-----------------------------------------
u32 getExternIp(void)
{
	return ip;
}
//-----------------------------------------
u32 getInternIp(void)
{
	if (int_ip) return int_ip;
	IP_ADAPTER_INFO add[20];
	ULONG size = sizeof(add);
	ULONG res = GetAdaptersInfo(add,&size);
	if (ERROR_SUCCESS == res)
	{
		for (IP_ADAPTER_INFO* ad = add; ad; ad = ad->Next)
		{
			for (IP_ADDR_STRING* st = &ad->IpAddressList; st; st = st->Next)
			{
				u32 ad_ip, ad_mask;
				u32 part[4];
				if (4 == sscanf_s(st->IpAddress.String,"%d.%d.%d.%d",&part[0],&part[1],&part[2],&part[3]))
				{
					ad_ip = (part[0]<<24)+(part[1]<<16)+(part[2]<<8)+part[3];
				}
				if (4 == sscanf_s(st->IpMask.String,"%d.%d.%d.%d",&part[0],&part[1],&part[2],&part[3]))
				{
					ad_mask = (part[0]<<24)+(part[1]<<16)+(part[2]<<8)+part[3];
				}
				if (int_srv_ip)
				{
					if (ad_ip && ad_mask && ((ad_ip&ad_mask) == (int_srv_ip&ad_mask)))
					{
						int_ip = ad_ip;
						break;
					}
				}
				else if (ad_ip && ad_mask)
				{
					int_ip = ad_ip;
					break;
				}
			}
		}
	}
	return int_ip;
}
//-----------------------------------------
bool setupUPnP(settings_t* set, int remove)
{
	struct UPNPDev * devlist = 0;
	char lanaddr[64];	// my ip address on the LAN
	int retcode = 0;
	int error = 0;
	const char proto[] = "UDP";

	if(devlist = upnpDiscover(2000, NULL,NULL,0,0,&error))
	{
		struct UPNPUrls urls;
		struct IGDdatas data;
		if(UPNP_GetValidIGD(devlist, &urls, &data, lanaddr, sizeof(lanaddr)) > 0)
		{
			char externalIPAddress[40];
			char eport[8];
			char iport[8];

			UPNP_GetExternalIPAddress(urls.controlURL,data.first.servicetype,externalIPAddress);

			u32 part[4];
			if (4 == sscanf_s(externalIPAddress,"%d.%d.%d.%d",&part[0],&part[1],&part[2],&part[3]))
			{
				ip = (part[0]<<24)+(part[1]<<16)+(part[2]<<8)+part[3];
			}
			if (4 == sscanf_s(urls.ipcondescURL,"http://%d.%d.%d.%d",&part[0],&part[1],&part[2],&part[3]))
			{
				int_srv_ip = (part[0]<<24)+(part[1]<<16)+(part[2]<<8)+part[3];
			}

			sprintf_s(eport,"%d",set->remotePort);
			sprintf_s(iport,"%d",set->localPort);
			if (remove)
			{
				retcode = UPNP_DeletePortMapping(urls.controlURL, data.first.servicetype, eport, proto, externalIPAddress);
				if (retcode==UPNPCOMMAND_SUCCESS) LogMessage(module,TEXT("UPnP mapping was removed successfully"));
				else LogMessage(module,TEXT("UPnP mapping can't be removed"));
			}
			else
			{
				int cnt = 0; // attempts counter
				do
				{
					if (set->randomPort)
					{
						cheap_random_number((u8*)&set->remotePort,sizeof(set->remotePort));
						sprintf_s(eport,"%d",set->remotePort);
					}
					retcode = UPNP_AddPortMapping(urls.controlURL,data.first.servicetype,eport,iport,lanaddr,"ZAS",proto,externalIPAddress,NULL);
					if (retcode==UPNPCOMMAND_SUCCESS) LogMessage(module,TEXT("UPnP mapping was set successfully"));
					else LogMessage(module,TEXT("ERROR: UPnP mapping can't be set"));
					if ((retcode!=UPNPCOMMAND_SUCCESS) && set->nextAvailablePort)
					{
						set->remotePort++;
						sprintf_s(eport,"%d",set->remotePort);
					}
					cnt++;
				} while ((retcode!=UPNPCOMMAND_SUCCESS)&&(set->nextAvailablePort||set->randomPort)&&(cnt>20));
			}
			FreeUPNPUrls(&urls);
		}
		else
		{
			LogMessage(module,TEXT("ERROR: UPnP mapping can't be set"));
		}
		freeUPNPDevlist(devlist);
		return (retcode==UPNPCOMMAND_SUCCESS);
	}
	LogMessage(module,TEXT("ERROR: UPnP mapping can't be set"));
	return false;
}
//------------------------------------------------------------------------
// WINDOWS FIREWALL ENABLE AND DISABLE
#include <windows.h>
#include <crtdbg.h>
#include <netfw.h>
#include <objbase.h>
#include <oleauto.h>
#include <stdio.h>

#pragma comment( lib, "ole32.lib" )
#pragma comment( lib, "oleaut32.lib" )

HRESULT WindowsFirewallInitialize(OUT INetFwProfile** fwProfile)
{
    HRESULT hr = S_OK;
    INetFwMgr* fwMgr = NULL;
    INetFwPolicy* fwPolicy = NULL;

    _ASSERT(fwProfile != NULL);

    *fwProfile = NULL;

    // Create an instance of the firewall settings manager.
    hr = CoCreateInstance(
            __uuidof(NetFwMgr),
            NULL,
            CLSCTX_INPROC_SERVER,
            __uuidof(INetFwMgr),
            (void**)&fwMgr
            );
    if (FAILED(hr))
    {
        printf("CoCreateInstance failed: 0x%08lx\n", hr);
        goto error;
    }

    // Retrieve the local firewall policy.
    hr = fwMgr->get_LocalPolicy(&fwPolicy);
    if (FAILED(hr))
    {
        printf("get_LocalPolicy failed: 0x%08lx\n", hr);
        goto error;
    }

    // Retrieve the firewall profile currently in effect.
    hr = fwPolicy->get_CurrentProfile(fwProfile);
    if (FAILED(hr))
    {
        printf("get_CurrentProfile failed: 0x%08lx\n", hr);
        goto error;
    }

error:

    // Release the local firewall policy.
    if (fwPolicy != NULL)
    {
        fwPolicy->Release();
    }

    // Release the firewall settings manager.
    if (fwMgr != NULL)
    {
        fwMgr->Release();
    }

    return hr;
}


void WindowsFirewallCleanup(IN INetFwProfile* fwProfile)
{
    // Release the firewall profile.
    if (fwProfile != NULL)
    {
        fwProfile->Release();
    }
}


HRESULT WindowsFirewallPortIsEnabled(
            IN INetFwProfile* fwProfile,
            IN LONG portNumber,
            IN NET_FW_IP_PROTOCOL ipProtocol,
            OUT BOOL* fwPortEnabled
            )
{
    HRESULT hr = S_OK;
    VARIANT_BOOL fwEnabled;
    INetFwOpenPort* fwOpenPort = NULL;
    INetFwOpenPorts* fwOpenPorts = NULL;

    _ASSERT(fwProfile != NULL);
    _ASSERT(fwPortEnabled != NULL);

    *fwPortEnabled = FALSE;

    // Retrieve the globally open ports collection.
    hr = fwProfile->get_GloballyOpenPorts(&fwOpenPorts);
    if (FAILED(hr))
    {
        printf("get_GloballyOpenPorts failed: 0x%08lx\n", hr);
        goto error;
    }

    // Attempt to retrieve the globally open port.
    hr = fwOpenPorts->Item(portNumber, ipProtocol, &fwOpenPort);
    if (SUCCEEDED(hr))
    {
        // Find out if the globally open port is enabled.
        hr = fwOpenPort->get_Enabled(&fwEnabled);
        if (FAILED(hr))
        {
            printf("get_Enabled failed: 0x%08lx\n", hr);
            goto error;
        }

        if (fwEnabled != VARIANT_FALSE)
        {
            // The globally open port is enabled.
            *fwPortEnabled = TRUE;

            printf("Port %ld is open in the firewall.\n", portNumber);
        }
        else
        {
            printf("Port %ld is not open in the firewall.\n", portNumber);
        }
    }
    else
    {
        // The globally open port was not in the collection.
        hr = S_OK;

        printf("Port %ld is not open in the firewall.\n", portNumber);
    }

error:

    // Release the globally open port.
    if (fwOpenPort != NULL)
    {
        fwOpenPort->Release();
    }

    // Release the globally open ports collection.
    if (fwOpenPorts != NULL)
    {
        fwOpenPorts->Release();
    }

    return hr;
}

HRESULT WindowsFirewallPortAdd(
            IN INetFwProfile* fwProfile,
            IN LONG portNumber,
            IN NET_FW_IP_PROTOCOL ipProtocol,
            IN const wchar_t* name
            )
{
    HRESULT hr = S_OK;
    BOOL fwPortEnabled;
    BSTR fwBstrName = NULL;
    INetFwOpenPort* fwOpenPort = NULL;
    INetFwOpenPorts* fwOpenPorts = NULL;

    _ASSERT(fwProfile != NULL);
    _ASSERT(name != NULL);

    // First check to see if the port is already added.
    hr = WindowsFirewallPortIsEnabled(
            fwProfile,
            portNumber,
            ipProtocol,
            &fwPortEnabled
            );
    if (FAILED(hr))
    {
        printf("WindowsFirewallPortIsEnabled failed: 0x%08lx\n", hr);
        goto error;
    }

    // Only add the port if it isn't already added.
    if (!fwPortEnabled)
    {
        // Retrieve the collection of globally open ports.
        hr = fwProfile->get_GloballyOpenPorts(&fwOpenPorts);
        if (FAILED(hr))
        {
            printf("get_GloballyOpenPorts failed: 0x%08lx\n", hr);
            goto error;
        }

        // Create an instance of an open port.
        hr = CoCreateInstance(
                __uuidof(NetFwOpenPort),
                NULL,
                CLSCTX_INPROC_SERVER,
                __uuidof(INetFwOpenPort),
                (void**)&fwOpenPort
                );
        if (FAILED(hr))
        {
            printf("CoCreateInstance failed: 0x%08lx\n", hr);
            goto error;
        }

        // Set the port number.
        hr = fwOpenPort->put_Port(portNumber);
        if (FAILED(hr))
        {
            printf("put_Port failed: 0x%08lx\n", hr);
            goto error;
        }

        // Set the IP protocol.
        hr = fwOpenPort->put_Protocol(ipProtocol);
        if (FAILED(hr))
        {
            printf("put_Protocol failed: 0x%08lx\n", hr);
            goto error;
        }

        // Allocate a BSTR for the friendly name of the port.
        fwBstrName = SysAllocString(name);
        if (SysStringLen(fwBstrName) == 0)
        {
            hr = E_OUTOFMEMORY;
            printf("SysAllocString failed: 0x%08lx\n", hr);
            goto error;
        }

        // Set the friendly name of the port.
        hr = fwOpenPort->put_Name(fwBstrName);
        if (FAILED(hr))
        {
            printf("put_Name failed: 0x%08lx\n", hr);
            goto error;
        }

        // Opens the port and adds it to the collection.
        hr = fwOpenPorts->Add(fwOpenPort);
        if (FAILED(hr))
        {
            printf("Add failed: 0x%08lx\n", hr);
            goto error;
        }

        printf("Port %ld is now open in the firewall.\n", portNumber);
    }

error:

    // Free the BSTR.
    SysFreeString(fwBstrName);

    // Release the open port instance.
    if (fwOpenPort != NULL)
    {
        fwOpenPort->Release();
    }

    // Release the globally open ports collection.
    if (fwOpenPorts != NULL)
    {
        fwOpenPorts->Release();
    }

    return hr;
}


HRESULT WindowsFirewallPortRemove(
            IN INetFwProfile* fwProfile,
            IN LONG portNumber,
            IN NET_FW_IP_PROTOCOL ipProtocol
            )
{
    HRESULT hr = S_OK;
    BOOL fwPortEnabled;
    INetFwOpenPorts* fwOpenPorts = NULL;

    _ASSERT(fwProfile != NULL);

    // First check to see if the port is already added.
    hr = WindowsFirewallPortIsEnabled(
            fwProfile,
            portNumber,
            ipProtocol,
            &fwPortEnabled
            );
    if (FAILED(hr))
    {
        printf("WindowsFirewallPortIsEnabled failed: 0x%08lx\n", hr);
        goto error;
    }

    // Only remove the port if it isn't already added.
    if (fwPortEnabled)
    {
        // Retrieve the collection of globally open ports.
        hr = fwProfile->get_GloballyOpenPorts(&fwOpenPorts);
        if (FAILED(hr))
        {
            printf("get_GloballyOpenPorts failed: 0x%08lx\n", hr);
            goto error;
        }

		fwOpenPorts->Remove(portNumber,ipProtocol);
    }

error:

    // Release the globally open ports collection.
    if (fwOpenPorts != NULL)
    {
        fwOpenPorts->Release();
    }

    return hr;
}

void EnablePortInFirewall(u16 port)
{
    HRESULT hr = S_OK;
    HRESULT comInit = E_FAIL;
    INetFwProfile* fwProfile = NULL;

    // Initialize COM.
    comInit = CoInitializeEx(0,COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

   // Ignore RPC_E_CHANGED_MODE; this just means that COM has already been
   // initialized with a different mode. Since we don't care what the mode is,
   // we'll just use the existing mode.
   if (comInit != RPC_E_CHANGED_MODE)
   {
        hr = comInit;
        if (FAILED(hr))
        {
            printf("CoInitializeEx failed: 0x%08lx\n", hr);
            goto error;
        }
   }

    // Retrieve the firewall profile currently in effect.
    hr = WindowsFirewallInitialize(&fwProfile);
    if (FAILED(hr))
    {
        printf("WindowsFirewallInitialize failed: 0x%08lx\n", hr);
        goto error;
    }

    hr = WindowsFirewallPortAdd(fwProfile, port, NET_FW_IP_PROTOCOL_UDP, TEXT("ZAS"));
    if (FAILED(hr))
    {
        printf("WindowsFirewallPortAdd failed: 0x%08lx\n", hr);
        goto error;
    }

error:

    // Release the firewall profile.
    WindowsFirewallCleanup(fwProfile);

    // Uninitialize COM.
    if (SUCCEEDED(comInit))
    {
        CoUninitialize();
    }
}

void DisablePortInFirewall(u16 port)
{
    HRESULT hr = S_OK;
    HRESULT comInit = E_FAIL;
    INetFwProfile* fwProfile = NULL;

    // Initialize COM.
    comInit = CoInitializeEx(0,COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

   // Ignore RPC_E_CHANGED_MODE; this just means that COM has already been
   // initialized with a different mode. Since we don't care what the mode is,
   // we'll just use the existing mode.
   if (comInit != RPC_E_CHANGED_MODE)
   {
        hr = comInit;
        if (FAILED(hr))
        {
            printf("CoInitializeEx failed: 0x%08lx\n", hr);
            goto error;
        }
   }

    // Retrieve the firewall profile currently in effect.
    hr = WindowsFirewallInitialize(&fwProfile);
    if (FAILED(hr))
    {
        printf("WindowsFirewallInitialize failed: 0x%08lx\n", hr);
        goto error;
    }

    hr = WindowsFirewallPortRemove(fwProfile, port, NET_FW_IP_PROTOCOL_UDP);
    if (FAILED(hr))
    {
        printf("WindowsFirewallPortAdd failed: 0x%08lx\n", hr);
        goto error;
    }

error:

    // Release the firewall profile.
    WindowsFirewallCleanup(fwProfile);

    // Uninitialize COM.
    if (SUCCEEDED(comInit))
    {
        CoUninitialize();
    }
}
