#include <pjmedia.h>
#include <pjlib.h>
#include <pjmedia-codec.h>
#include <pjmedia-videodev/videodev.h>
#include <pjmedia/vid_codec.h>
#include <pjmedia/port.h>
#include <winsock2.h>
#include <windows.h>
#include "network.h"
#include "control.h"
#include "main.h"

/* 
 * Capture device setting: 
 *   -1 = colorbar, 
 *   -2 = any non-colorbar capture device (first found)
 *    x = specified capture device id
 */
#define MAX_PACKETS		15
static const u16 NETWORK_SIZE_VIDEO_MAX = NETWORK_SIZE_USEFUL-8;
static const u32 MAX_PACKET_SIZE = 255*(u32)NETWORK_SIZE_VIDEO_MAX;

typedef struct {
	pj_pool_t			*video_pool;
	pjmedia_port		codec_port;
    pjmedia_vid_codec_param codec_param;
	pjmedia_vid_port_param vport_param;
    pjmedia_vid_codec   *codec;
    pjmedia_vid_port    *renderer;
    pjmedia_vid_port    *preview;
	pjmedia_vid_port	*capture;
	pj_uint8_t			*tx_data;
	pj_size_t			 tx_data_size;
	pj_uint8_t			*rx_data;
	pj_size_t			 rx_data_size;
	NODE				 nd;
	u16					 rx_cnt;
	u16					 tx_cnt;
	pjmedia_frame		 rx_frames[MAX_PACKETS];

} vid_data_t;

static pj_caching_pool caching_pool;
static pj_pool_t *pool;

static pj_status_t codec_on_event(pjmedia_event *event, void *user_data)
{
    vid_data_t *vid_data = (vid_data_t*)user_data;

    if (event->type == PJMEDIA_EVENT_FMT_CHANGED) {
		pj_status_t status;

		status = pjmedia_vid_codec_get_param(vid_data->codec, &vid_data->codec_param);
		if (status != PJ_SUCCESS)
			return status;

		status = pjmedia_vid_dev_stream_set_cap(pjmedia_vid_port_get_stream(vid_data->renderer),PJMEDIA_VID_DEV_CAP_FORMAT,&vid_data->codec_param.dec_fmt);
		if (status != PJ_SUCCESS)
			return status;

		status = pjmedia_vid_dev_stream_set_cap(pjmedia_vid_port_get_stream(vid_data->renderer),PJMEDIA_VID_DEV_CAP_OUTPUT_RESIZE,&vid_data->codec_param.dec_fmt.det.vid.size);
		if (status != PJ_SUCCESS)
			return status;
    }

    return PJ_SUCCESS;
}

static pj_status_t codec_show_frames(vid_data_t *vid_data, pjmedia_frame *frame, unsigned num, char* name)
{
    pj_status_t status;
    pjmedia_vid_codec *codec = vid_data->codec;
	pjmedia_frame out_frame;
	out_frame.buf = vid_data->rx_data;

	if(!vid_data->renderer){
		/* Create renderer, set it to passive (slave)  */
	    pjmedia_vid_port_param_default(&vid_data->vport_param);
		pjmedia_vid_dev_default_param(vid_data->video_pool, 0, &vid_data->vport_param.vidparam);
		pjmedia_format_copy(&vid_data->vport_param.vidparam.fmt, &vid_data->codec_param.dec_fmt);
		vid_data->vport_param.vidparam.dir = PJMEDIA_DIR_CAPTURE;
		vid_data->vport_param.active = PJ_TRUE;
		pjmedia_video_format_detail *vfd = pjmedia_format_get_video_format_detail(&vid_data->vport_param.vidparam.fmt,PJ_TRUE);
		pjmedia_vid_dev_index rdr_idx;
		pjmedia_vid_dev_lookup("SDL", "SDL renderer", &rdr_idx);
		vid_data->vport_param.active = PJ_FALSE;
		vid_data->vport_param.vidparam.flags = PJMEDIA_VID_DEV_CAP_FORMAT|PJMEDIA_VID_DEV_CAP_OUTPUT_WINDOW_FLAGS|PJMEDIA_VID_DEV_CAP_WINDOW_TITLE;
		vid_data->vport_param.vidparam.dir = PJMEDIA_DIR_RENDER;
	    vid_data->vport_param.vidparam.disp_size = vfd->size;
		vid_data->vport_param.vidparam.rend_id = rdr_idx;
		vid_data->vport_param.vidparam.window_flags = PJMEDIA_VID_DEV_WND_RESIZABLE|PJMEDIA_VID_DEV_WND_BORDER;
		vid_data->vport_param.vidparam.title = name;

		pjmedia_vid_port_create(vid_data->video_pool, &vid_data->vport_param, &vid_data->renderer);
		pjmedia_vid_port_start(vid_data->renderer);
    }

    // Decode
    status = pjmedia_vid_codec_decode(codec, num, frame,(unsigned)vid_data->rx_data_size, &out_frame);
    if (status != PJ_SUCCESS) 
		goto on_error;

    /* Display */
	status = pjmedia_port_put_frame(pjmedia_vid_port_get_passive_port(vid_data->renderer),&out_frame);
    if (status != PJ_SUCCESS)
		goto on_error;

    return PJ_SUCCESS;

on_error:
    return status;
}

static void send_video_frames(vid_data_t *vid_data, pjmedia_frame* frame, u32 num)
{
	if (isZeroID(vid_data->nd.id)) return;
	u8 buf[NETWORK_SIZE_USEFUL] = {0};
	buf[0] = CMD_VIDEO;
	*(u16*)&buf[1] = ++vid_data->tx_cnt;
	*(u16*)&buf[1+2] = NETWORK_SIZE_VIDEO_MAX;
	for (u32 n = 0; n < num; n++){
		buf[1+2+2] = (u8)((n|(num<<4))&0xFF);
		buf[1+2+2+1] = (u8)((frame[n].size+NETWORK_SIZE_VIDEO_MAX-1)/NETWORK_SIZE_VIDEO_MAX);
		for (u32 sent = 0; sent < frame[n].size;)
		{
			u32 len = frame[n].size-sent;
			if (len > NETWORK_SIZE_VIDEO_MAX) len = NETWORK_SIZE_VIDEO_MAX;
			buf[1+2+2+2] = (u8)(sent/NETWORK_SIZE_VIDEO_MAX);
			CopyMemory(&buf[8],(u8*)frame[n].buf+sent,len);
			PACKET_FOR_SEND pack = {0};
			pack.buffer = buf;
			pack.buf_len = 8 + len;
			pack.nd = &vid_data->nd;
			pack.protocol = NETWORK_PROT_CONTROL;
			pack.dont_count_tx=true;
			getNet()->send_packet(&pack);
			sent += len;
		}
	}
}

void video_put_data(void* vid,u8* buf, u32 buf_size, char* name)
{
	if (!vid) return;
	vid_data_t* vid_data = (vid_data_t*)vid;

	if ((buf_size > 8) && (buf[0] == CMD_VIDEO)){
		u16 rx_num = *(u16*)&buf[1];
		u16 chunk_size = *(u16*)&buf[1+2];
		u8 frame_total = buf[1+2+2]>>4;
		u8 frame_num = buf[1+2+2]&0x0F;
		u8 pack_total = buf[1+2+2+1];
		u8 pack_num = buf[1+2+2+2];
		if (rx_num != vid_data->rx_cnt){ // clear data for new set of frames
			vid_data->rx_cnt = rx_num;
			for (int cnt = 0; cnt < MAX_PACKETS; cnt++) {
				vid_data->rx_frames[cnt].type = PJMEDIA_FRAME_TYPE_NONE;
				vid_data->rx_frames[cnt].size = 0;
			}
			vid_data->rx_frames[0].buf = vid_data->rx_data;
		}
		if (frame_total && frame_total < MAX_PACKETS && pack_total && frame_num < frame_total && pack_num < pack_total && (!frame_num || (vid_data->rx_frames[frame_num-1].type == PJMEDIA_FRAME_TYPE_VIDEO))){
			u8* addr = (u8*)vid_data->rx_frames[frame_num].buf + (u32)pack_num*chunk_size;
			if ((u32)(addr + buf_size-8 - vid_data->rx_data) <= vid_data->rx_data_size){
				memcpy(addr,&buf[8],buf_size-8);
			}
			if (pack_num == pack_total-1){ // last packet for this frame
				vid_data->rx_frames[frame_num].size = (u32)pack_num*chunk_size+buf_size-8;
				vid_data->rx_frames[frame_num].type = PJMEDIA_FRAME_TYPE_VIDEO;
				if (frame_num < MAX_PACKETS-1) vid_data->rx_frames[frame_num+1].buf = (u8*)vid_data->rx_frames[frame_num].buf + vid_data->rx_frames[frame_num].size;
				if (frame_num == frame_total-1) { // last frame, show it
					codec_show_frames(vid_data, vid_data->rx_frames, frame_total, name);
				}
			}
		}
	}
}

static pj_status_t codec_put_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    vid_data_t *vid_data = (vid_data_t*)port->port_data.pdata;
    pj_status_t status;
    pjmedia_vid_codec *codec = vid_data->codec;
    unsigned enc_cnt = 0;
    pj_uint8_t *enc_buf;
    unsigned enc_size_left;
    pjmedia_frame enc_frames[MAX_PACKETS];
    pj_bool_t has_more = PJ_FALSE;

    status = pjmedia_port_put_frame(pjmedia_vid_port_get_passive_port(vid_data->preview),frame);
    if (status != PJ_SUCCESS)
		goto on_error;

    enc_buf = vid_data->tx_data;
    enc_size_left = (unsigned)vid_data->tx_data_size;

    // Encode
    enc_frames[enc_cnt].buf = enc_buf;
    enc_frames[enc_cnt].size = enc_size_left;

    status = pjmedia_vid_codec_encode_begin(codec, NULL, frame, enc_size_left, &enc_frames[enc_cnt], &has_more);
    if (status != PJ_SUCCESS) 
		goto on_error;

    enc_buf += enc_frames[enc_cnt].size;
    enc_size_left -= (unsigned)enc_frames[enc_cnt].size;
	if (enc_frames[enc_cnt].size == 0)
		return PJ_SUCCESS;

    ++enc_cnt;
    while (has_more) {
		enc_frames[enc_cnt].buf = enc_buf;
		enc_frames[enc_cnt].size = enc_size_left;

		status = pjmedia_vid_codec_encode_more(codec, enc_size_left,&enc_frames[enc_cnt],&has_more);
		if (status != PJ_SUCCESS)
			break;

		enc_buf += enc_frames[enc_cnt].size;
		enc_size_left -= (unsigned)enc_frames[enc_cnt].size;

		++enc_cnt;

		if (enc_cnt >= MAX_PACKETS) {
			assert(!"Too many packets!");
			break;
		}
    }

	// Send frame
	if (enc_frames[0].type == PJMEDIA_FRAME_TYPE_VIDEO)
		//codec_show_frames(vid_data, enc_frames,enc_cnt);
		send_video_frames(vid_data,enc_frames,enc_cnt);

    return PJ_SUCCESS;

on_error:
    return status;
}

void video_delete(void* vid)
{
	if (!vid) return;
	vid_data_t* vid_data = (vid_data_t*)vid;
    if (vid_data->capture) pjmedia_vid_port_stop(vid_data->capture);
    if (vid_data->renderer) pjmedia_vid_port_stop(vid_data->renderer);
    if (vid_data->preview) pjmedia_vid_port_stop(vid_data->preview);
    if (vid_data->capture) pjmedia_vid_port_destroy(vid_data->capture);
    if (vid_data->renderer) pjmedia_vid_port_destroy(vid_data->renderer);
    if (vid_data->preview) pjmedia_vid_port_destroy(vid_data->preview);
	pjmedia_event_unsubscribe(NULL, &codec_on_event, &vid_data, vid_data->codec);
    pjmedia_vid_codec_close(vid_data->codec);
    pjmedia_vid_codec_mgr_dealloc_codec(NULL, vid_data->codec);
	pj_pool_release(vid_data->video_pool);
	delete vid_data;
}

void* video_create(void)
{
	vid_data_t* vid_data = new vid_data_t();
    const pj_str_t port_name = {"codec", 5};
    pjmedia_vid_codec *codec=NULL;
    const pjmedia_vid_codec_info *codec_info;
    pj_status_t status;

    pj_str_t codec_id_st;
    unsigned info_cnt = 1;

	vid_data->video_pool = pj_pool_create(&caching_pool.factory, "vid_data mem", 4000, 1024, NULL);

    /* Lookup codec */
    pj_cstr(&codec_id_st, "h264");
    status = pjmedia_vid_codec_mgr_find_codecs_by_id(NULL, &codec_id_st, &info_cnt, &codec_info, NULL);
    if (status != PJ_SUCCESS) {
        goto on_return;
    }

    status = pjmedia_vid_codec_mgr_get_default_param(NULL, codec_info, &vid_data->codec_param);
    if (status != PJ_SUCCESS) {
        goto on_return;
    }

    vid_data->codec_param.packing = PJMEDIA_VID_PACKING_PACKETS;

    /* Don't apply SDP fmtp */
    vid_data->codec_param.ignore_fmtp = PJ_TRUE;
	vid_data->codec_param.enc_mtu = MAX_PACKET_SIZE;
	switch(getVQ()){
	case 1: // normal
		vid_data->codec_param.dec_fmt.det.vid.avg_bps = vid_data->codec_param.dec_fmt.det.vid.max_bps = 128000;
		vid_data->codec_param.enc_fmt.det.vid.avg_bps = vid_data->codec_param.enc_fmt.det.vid.max_bps = 128000;
		break;
	case 2: // good
		vid_data->codec_param.dec_fmt.det.vid.avg_bps = vid_data->codec_param.dec_fmt.det.vid.max_bps = 256000;
		vid_data->codec_param.enc_fmt.det.vid.avg_bps = vid_data->codec_param.enc_fmt.det.vid.max_bps = 256000;
		break;
	default: // low
		vid_data->codec_param.dec_fmt.det.vid.avg_bps = vid_data->codec_param.dec_fmt.det.vid.max_bps = 64000;
		vid_data->codec_param.enc_fmt.det.vid.avg_bps = vid_data->codec_param.enc_fmt.det.vid.max_bps = 64000;
		break;
	}
	switch (getVR()){
	case 0: 
		vid_data->codec_param.enc_fmt.det.vid.size.w = vid_data->codec_param.dec_fmt.det.vid.size.w = 480;
		vid_data->codec_param.enc_fmt.det.vid.size.h = vid_data->codec_param.dec_fmt.det.vid.size.h = 320;
		break;
	case 2: 
		vid_data->codec_param.enc_fmt.det.vid.size.w = vid_data->codec_param.dec_fmt.det.vid.size.w = 800;
		vid_data->codec_param.enc_fmt.det.vid.size.h = vid_data->codec_param.dec_fmt.det.vid.size.h = 600;
		break;
	case 3: 
		vid_data->codec_param.enc_fmt.det.vid.size.w = vid_data->codec_param.dec_fmt.det.vid.size.w = 1280;
		vid_data->codec_param.enc_fmt.det.vid.size.h = vid_data->codec_param.dec_fmt.det.vid.size.h = 720;
		break;
	default: 
		vid_data->codec_param.enc_fmt.det.vid.size.w = vid_data->codec_param.dec_fmt.det.vid.size.w = 720;
		vid_data->codec_param.enc_fmt.det.vid.size.h = vid_data->codec_param.dec_fmt.det.vid.size.h = 480;
		break;
	}

    /* Open codec */
    status = pjmedia_vid_codec_mgr_alloc_codec(NULL, codec_info,&codec);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

    status = pjmedia_vid_codec_init(codec, vid_data->video_pool);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

    status = pjmedia_vid_codec_open(codec, &vid_data->codec_param);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

	/* After opened, codec will update the param, let's sync encoder & 
	 * decoder format detail.
	 */
	vid_data->codec_param.dec_fmt.det = vid_data->codec_param.enc_fmt.det;

	/* Subscribe to codec events */
	pjmedia_event_subscribe(NULL, &codec_on_event, vid_data, codec);


    /* Init codec port */
    pj_bzero(&vid_data->codec_port, sizeof(vid_data->codec_port));
    status = pjmedia_port_info_init2(&vid_data->codec_port.info, &port_name, 0x1234,PJMEDIA_DIR_ENCODING, &vid_data->codec_param.dec_fmt);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

    vid_data->codec = codec;
    vid_data->rx_data_size = 1920 * 1080 * 4;
    vid_data->rx_data = (pj_uint8_t*)pj_pool_alloc(vid_data->video_pool, vid_data->rx_data_size);
    vid_data->tx_data_size = vid_data->rx_data_size;
    vid_data->tx_data = (pj_uint8_t*)pj_pool_alloc(vid_data->video_pool, vid_data->tx_data_size);
    
    vid_data->codec_port.put_frame = &codec_put_frame;
    vid_data->codec_port.port_data.pdata = vid_data;

	return vid_data;
on_return:
	video_delete(vid_data);
	return NULL;
}

void video_stop(void* vid)
{
	if (!vid) return;
	vid_data_t* vid_data = (vid_data_t*)vid;
    if (vid_data->capture) pjmedia_vid_port_stop(vid_data->capture);
    if (vid_data->preview) pjmedia_vid_port_stop(vid_data->preview);
    if (vid_data->capture) pjmedia_vid_port_destroy(vid_data->capture);
    if (vid_data->preview) pjmedia_vid_port_destroy(vid_data->preview);
    vid_data->capture = NULL;
    vid_data->preview = NULL;
}

void video_stream_finished(void* vid)
{
	if (!vid) return;
	vid_data_t* vid_data = (vid_data_t*)vid;
    if (vid_data->renderer) pjmedia_vid_port_stop(vid_data->renderer);
    if (vid_data->renderer) pjmedia_vid_port_destroy(vid_data->renderer);
    vid_data->renderer = NULL;
}

bool video_start(void* vid, NODE* nd, LPCTSTR vid_str)
{
	unsigned i;
    pjmedia_vid_dev_index cap_idx, rdr_idx;
    pjmedia_video_format_detail *vfd;
	pj_status_t status;
	if (!vid) return false;
	vid_data_t* vid_data = (vid_data_t*)vid;

	if (nd)
		vid_data->nd = *nd;
	else
		ZeroMemory(&vid_data->nd,sizeof(NODE));

	// find video
	char vid_name[64];
#ifdef _UNICODE
	WideCharToMultiByte(CP_ACP,0,vid_str,64,vid_name,sizeof(vid_name),NULL,NULL);
#else
	strcpy(vid_name,getVideo());
#endif
	cap_idx = 0;
	for (i = 0; i < pjmedia_vid_dev_count(); ++i) {
		pjmedia_vid_dev_info info;
	    if (PJ_SUCCESS != pjmedia_vid_dev_get_info(i, &info)){
			continue;
	    }
	    if (info.dir & PJMEDIA_DIR_CAPTURE && !pj_ansi_stricmp(info.name, vid_name)){
		    cap_idx = i;
			break;
		}
	}

    pjmedia_vid_port_param_default(&vid_data->vport_param);
	/* Lookup SDL renderer */
	status = pjmedia_vid_dev_lookup("SDL", "SDL renderer", &rdr_idx);
	if (status != PJ_SUCCESS) {
		goto on_return;
	}

    /* Create capture, set it to active (master) */
    status = pjmedia_vid_dev_default_param(vid_data->video_pool, cap_idx, &vid_data->vport_param.vidparam);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }
    pjmedia_format_copy(&vid_data->vport_param.vidparam.fmt, &vid_data->codec_param.dec_fmt);
    vid_data->vport_param.vidparam.dir = PJMEDIA_DIR_CAPTURE;
    vid_data->vport_param.active = PJ_TRUE;

    if (vid_data->vport_param.vidparam.fmt.detail_type != PJMEDIA_FORMAT_DETAIL_VIDEO) {
		goto on_return;
    }

    vfd = pjmedia_format_get_video_format_detail(&vid_data->vport_param.vidparam.fmt,PJ_TRUE);
    if (vfd == NULL) {
		goto on_return;
    }

    status = pjmedia_vid_port_create(vid_data->video_pool, &vid_data->vport_param, &vid_data->capture);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

    /* Create preview, set it to passive (slave)  */
    vid_data->vport_param.active = PJ_FALSE;
	vid_data->vport_param.vidparam.flags = PJMEDIA_VID_DEV_CAP_OUTPUT_POSITION|PJMEDIA_VID_DEV_CAP_OUTPUT_RESIZE|PJMEDIA_VID_DEV_CAP_FORMAT|PJMEDIA_VID_DEV_CAP_OUTPUT_WINDOW_FLAGS|PJMEDIA_VID_DEV_CAP_WINDOW_TITLE;
    vid_data->vport_param.vidparam.dir = PJMEDIA_DIR_RENDER;
    vid_data->vport_param.vidparam.rend_id = rdr_idx;
    vid_data->vport_param.vidparam.disp_size = vfd->size;
	vid_data->vport_param.vidparam.disp_size.h = 240;
	vid_data->vport_param.vidparam.disp_size.w = 360;
	vid_data->vport_param.vidparam.window_pos.x = 5;
	vid_data->vport_param.vidparam.window_pos.y = 30;
	vid_data->vport_param.vidparam.window_flags = PJMEDIA_VID_DEV_WND_RESIZABLE|PJMEDIA_VID_DEV_WND_BORDER;
	vid_data->vport_param.vidparam.title = "Preview";

	status = pjmedia_vid_port_create(vid_data->video_pool, &vid_data->vport_param, &vid_data->preview);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }
    /* Connect capture to codec port */
    status = pjmedia_vid_port_connect(vid_data->capture, &vid_data->codec_port, PJ_FALSE);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

    /* Start streaming.. */
    status = pjmedia_vid_port_start(vid_data->preview);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }
    status = pjmedia_vid_port_start(vid_data->capture);
    if (status != PJ_SUCCESS) {
		goto on_return;
    }

	return true;
on_return:
    video_stop(vid_data);

	return false;
}

void video_init(void)
{
	pj_pool_factory *mem;

    pj_init();
    pj_caching_pool_init(&caching_pool, &pj_pool_factory_default_policy, 0);
    pool = pj_pool_create(&caching_pool.factory, "video", 1000, 512, NULL);
    mem = &caching_pool.factory;

    pjmedia_video_format_mgr_create(pool, 64, 0, NULL);
    pjmedia_converter_mgr_create(pool, NULL);
    pjmedia_event_mgr_create(pool, 0, NULL);
    pjmedia_vid_codec_mgr_create(pool, NULL);
    pjmedia_vid_dev_subsys_init(mem);
    pjmedia_codec_openh264_vid_init(NULL, mem);
}

void video_destroy(void)
{
    pjmedia_codec_openh264_vid_deinit();
    pjmedia_vid_dev_subsys_shutdown();
    pjmedia_video_format_mgr_destroy(pjmedia_video_format_mgr_instance());
    pjmedia_converter_mgr_destroy(pjmedia_converter_mgr_instance());
    pjmedia_event_mgr_destroy(pjmedia_event_mgr_instance());
    pjmedia_vid_codec_mgr_destroy(pjmedia_vid_codec_mgr_instance());

	pj_pool_release(pool);
    pj_caching_pool_destroy(&caching_pool);
}

void fill_video_inputs(HWND cb)
{
	for (unsigned i = 0; i < pjmedia_vid_dev_count(); ++i) {
		pjmedia_vid_dev_info info;
	    if (PJ_SUCCESS != pjmedia_vid_dev_get_info(i, &info)){
			continue;
	    }
	    if (info.dir & PJMEDIA_DIR_CAPTURE /*&& pj_ansi_stricmp(info.driver, "Colorbar")*/)
	    {
			TCHAR name[sizeof(info.name)];
#ifdef _UNICODE
			MultiByteToWideChar(CP_ACP,0,info.name,sizeof(info.name),name,sizeof(info.name));
#else
			strcpy(name,info.name);
#endif
			ComboBox_AddString(cb,name);
	    }
	}
}
